#
# Main package initializer
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import subprocess
import warnings
import sys
from importlib.metadata import version, PackageNotFoundError
from typing import List

# Get package version: either via meta-information from egg or via latest git commit
try:
    __version__ = version("esi-acme")
except PackageNotFoundError:
    proc = subprocess.Popen("git describe --always",
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            text=True, shell=True)
    out, err = proc.communicate()
    if proc.returncode != 0:                                            # pragma: no cover
        proc = subprocess.Popen("git rev-parse HEAD:acme/__init__.py",
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                text=True, shell=True)
        out, err = proc.communicate()
        if proc.returncode != 0:
            msg = "<ACME> Package is not installed in site-packages nor cloned via git. " +\
                "Please consider obtaining ACME sources from supported channels. "
            warnings.warn(msg)
            out = "-999"
    __version__ = out.rstrip("\n")

# Central collection of deprecated keywords/canonical warning message
__deprecated__ = ["n_jobs", "mem_per_job", "n_jobs_startup" "workers_per_job"]
__deprecation_wrng__ = \
    "The keywords `n_jobs`, `mem_per_job`, `n_jobs_startup` and " +\
    "`workers_per_job` are DEPRECATED. Please use `n_workers`, `mem_per_worker`, " +\
    "`n_workers_startup` and `processes_per_worker`, respectively."

# Remove dask-jobqueue's FutureWarnings about tmpfile (which we don't use)
warnings.simplefilter(action='ignore', category=FutureWarning)

# Import local modules
from . import frontend, backend, shared, dask_helpers
from .frontend import *
from .backend import *
from .shared import *
from .dask_helpers import *
from .logger import prepare_log

# Set up module-wide logging
prepare_log(logname="ACME")

# Override default exception handler (take care of Jupyter's Exception handling)
from .shared import ctrlc_catcher
try:
    ipy = get_ipython()                                                             # type: ignore
    import IPython
    ipy.ipyTBshower = IPython.core.interactiveshell.InteractiveShell.showtraceback
    IPython.core.interactiveshell.InteractiveShell.showtraceback = ctrlc_catcher    # type: ignore
except NameError:
    sys.excepthook = ctrlc_catcher

# Manage user-exposed namespace imports
__all__: List[str] = []
__all__.extend(frontend.__all__)
__all__.extend(backend.__all__)
__all__.extend(shared.__all__)
__all__.extend(dask_helpers.__all__)



#
# Auxiliaries needed across the entire package
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import os
import sys
import socket
import platform
import select
import subprocess
import inspect
import logging
import traceback
import numpy as np
import dask.distributed as dd
from logging import handlers
from typing import Any, Optional, List

# Local imports
from acme import __version__
from . import dask_helpers as dh

callCount = 0
callMax = 1000000

__all__: List["str"] = []


def sizeOf(
        obj: Any,
        varname: str) -> float:
    """
    Estimate memory consumption of Python objects

    Parameters
    ----------
    obj : Python object
        Any valid Python object whose memory footprint is of interest.
    varname : str
        Assigned name of `obj`

    Returns
    -------
    objsize : float
        Approximate memory footprint of `obj` in megabytes (MB).

    Notes
    -----
    Memory consumption is is estimated by recursively calling :meth:`sys.getsizeof`.
    Circular object references are followed up to a (preset) maximal recursion
    depth. This method was inspired by a routine in
    `Nifty <https://github.com/mwojnars/nifty/blob/master/util.py>`_.
    """

    # Keep track of the no. of recursive calls
    global callCount

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Protect against circular object references
    callCount += 1
    if callCount >= callMax:
        msg = "%s maximum recursion depth %s exceeded while processing %s"
        raise RecursionError(msg%(funcName, callMax, varname))

    # Use `sys.getsizeof` to estimate memory consumption of primitive objects
    objsize = sys.getsizeof(obj) / 1024**2
    if isinstance(obj, dict):
        return objsize + sum(list(map(sizeOf, obj.keys(), [varname] * len(obj.keys())))) + sum(list(map(sizeOf, obj.values(), [varname] * len(obj.values()))))
    if isinstance(obj, (list, tuple, set)):
        return objsize + sum(list(map(sizeOf, obj, [varname] * len(obj))))
    return objsize


def is_slurm_node() -> bool:
    """
    Returns `True` if code is running on a SLURM-managed cluster node, `False`
    otherwise
    """

    # Fetch ACME logger
    log = logging.getLogger("ACME")

    # Simply test if the srun command is available
    log.debug("Test if `sinfo` is available")
    out, _ = subprocess.Popen("sinfo --version",
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              text=True, shell=True).communicate()
    return len(out) > 0


def is_esi_node() -> bool:
    """
    Returns `True` if code is running on an ESI cluster node, `False` otherwise
    """

    # Fetch ACME logger and write debug message
    log = logging.getLogger("ACME")
    log.debug("Test if hostname matches the pattern 'esi-sv*'")
    return socket.gethostname().startswith("esi-sv") and os.path.isdir("/cs")


def is_x86_node() -> bool:
    """
    Returns `True` if code is running on an x86_64 node, `False` otherwise
    """

    # Fetch ACME logger and write debug message
    log = logging.getLogger("ACME")
    log.debug("Test if host is x86_64 micro-architecture")
    return platform.machine() == "x86_64"

def _scalar_parser(
        var: Any,
        varname: str = "varname",
        ntype: str = "int_like",
        lims: List = [-np.inf, np.inf]) -> None:
    """
    ACME-specific version of Syncopy's `scalar_parser` (used for cross-compatibility)
    """

    # Get name of calling method/function
    funcName = "<{}>".format(inspect.currentframe().f_back.f_code.co_name)      # type: ignore

    # Make sure `var` is a scalar-like number
    msg = "%s `%s` has to be %s between %s and %s, not %s"
    if np.issubdtype(type(var), np.number):
        error = False
        if ntype == "int_like":
            scalartype = "an integer"
            if round(var) != var:
                error = True
        else:
            scalartype = "a number"
        if var < lims[0] or var > lims[1]:
            error = True
        if error:
            raise ValueError(msg%(funcName,
                                  varname,
                                  scalartype,
                                  str(lims[0]),
                                  str(lims[1]),
                                  str(var)))
    else:
        msg = "%s `%s` has to be a scalar, not %s"
        raise TypeError(msg%(funcName, varname, str(type(var))))

    return


def user_yesno(                                                         # pragma: no cover
        msg: str,
        default: Optional[str] = None) -> bool:
    """
    ACME specific version of user-input query
    """

    # Parse optional `default` answer
    valid = {"yes": True, "y": True, "ye":True, "no":False, "n":False}
    if default is None:
        suffix = " [y/n] "
    elif default == "yes":
        suffix = " [Y/n] "
    elif default == "no":
        suffix = " [y/N] "

    # Wait for valid user input, if received return `True`/`False`
    while True:
        choice = input(msg + suffix).lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid.keys():
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")


def user_input(                                                         # pragma: no cover
        msg: str,
        valid: Optional[List] = None,
        default: Optional[str] = None,
        timeout: Optional[float] = None) -> str:
    """
    ACME specific version of user-input query
    """

    # Prepare to log any uncaught exceptions
    log = logging.getLogger("ACME")

    # Add trailing whitespace to `msg` if not already present and append
    # default reply (if provided)
    suffix = "" + " " * (not msg.endswith(" "))
    if default is not None:
        default = default.replace("[", "").replace("]","")
        if valid is not None:
            assert default in valid
        suffix = f"[Default: '{default}'] "
    query = msg + suffix

    # Jupyter only supports hard-blocking `input` fields
    if is_jupyter():
        log.debug("Running inside Jupyter notebook, deactivating timeout")
        timeout = None

    # Wait for user I/O
    print(query)
    while True:
        if timeout is None:
            choice = input()
        else:
            stdin, _, _ = select.select([sys.stdin], [], [], timeout)
            if stdin:
                choice = sys.stdin.readline().strip()
            else:
                if default is None:
                    err = f"No response received within the given timeout of {timeout} seconds. "
                    raise TimeoutError(err)
        if default is not None and choice == "":
            return default
        elif valid is not None and choice not in valid:
            print("Please respond with " + " or ".join(valid))
        else:
            return choice


def is_jupyter() -> bool:
    try:
        return get_ipython().__class__.__name__ == "ZMQInteractiveShell"    # type: ignore
    except NameError:
        return False


def ctrlc_catcher(
        *excargs: Any,
        **exckwargs: Optional[Any]) -> None:
    """
    Custom Traceback for properly handling CTRL + C interrupts while parallel
    computations are running
    """

    # Depending on the number of input arguments, we're either in Jupyter/iPython
    # or "regular" Python - this matters for actually handling the raised exception
    if len(excargs) == 3:
        isipy = False
        etype, evalue, etb = excargs
    else:
        shell, = excargs
        etype, evalue, etb = sys.exc_info()
        try:                            # careful: if iPython is used to launch a script, ``get_ipython`` is not defined
            get_ipython()               # type: ignore
            isipy = True
            sys.last_traceback = etb    # smartify ``sys``
        except NameError:
            isipy = False

    # Prepare to log any uncaught exceptions
    log = logging.getLogger("ACME")

    # The only exception we really care about is a `KeyboardInterrupt`: if CTRL + C
    # is pressed, ensure graceful shutdown of any parallel processing clients
    if issubclass(etype, KeyboardInterrupt):
        try:
            client = dd.get_client()
        except ValueError:
            client = None
        if client is not None:
            for st in client.futures.values():
                st.cancel()
            client.futures.clear()
            dh.cluster_cleanup(client)
            log.debug("CTRL + C acknowledged, client and workers successfully killed")

    # Relay exception handling back to appropriate system tools
    if isipy:                                                           # pragma: no cover
        shell.ipyTBshower(shell, exc_tuple=(etype, evalue, etb), **exckwargs)
    else:
        sys.__excepthook__(etype, evalue, etb)

    # Write to all logging locations, manually print traceback to file (stdout
    # printing was handled above)
    log.error("Exception received.")
    memHandler = [h for h in log.handlers if isinstance(h, handlers.MemoryHandler)][0]
    if memHandler.target is not None:                                   # pragma: no cover
        memHandler.acquire()
        with open(memHandler.target.baseFilename, "a", encoding="utf-8") as logfile:    # type: ignore
            logfile.write("".join(traceback.format_exception_only(etype, evalue)))
            logfile.write("".join(traceback.format_tb(etb)))
        memHandler.release()

    return



#
# Helper routines for working w/dask
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import os
import sys
import socket
import platform
import subprocess
import getpass
import time
import inspect
import textwrap
import numpy as np
from tqdm import tqdm
if sys.platform == "win32":                                             # pragma: no cover
    # tqdm breaks term colors on Windows - fix that (tqdm issue #446)
    import colorama
    colorama.deinit()
    colorama.init(strip=False)
from dask_jobqueue import SLURMCluster
from dask.distributed import Client, get_client, LocalCluster
from datetime import datetime, timedelta
from typing import List, Optional, Any, Union

# Local imports
from acme import __deprecated__, __deprecation_wrng__
from .shared import user_input, user_yesno, is_jupyter
from .spy_interface import scalar_parser, log

__all__: List["str"] = ["esi_cluster_setup", "local_cluster_setup", "cluster_cleanup", "slurm_cluster_setup"]


# Setup SLURM workers on the ESI HPC cluster
def esi_cluster_setup(
        partition: str,
        n_workers: int = 2,
        mem_per_worker: str = "auto",
        cores_per_worker: Optional[int] = None,
        n_workers_startup: int = 1,
        timeout: int = 60,
        interactive: bool = True,
        interactive_wait: int = 120,
        start_client: bool = True,
        job_extra: List = [],
        **kwargs: Optional[Any]) -> Union[Client, SLURMCluster, LocalCluster]:
    """
    Start a Dask distributed SLURM worker cluster on the ESI HPC infrastructure
    (or local multi-processing)

    Parameters
    ----------
    partition : str
        Name of SLURM partition/queue to start workers in. Use the command `sinfo`
        in the terminal to see a list of available SLURM partitions on the ESI HPC
        cluster.
    n_workers : int
        Number of SLURM workers (=jobs) to spawn
    mem_per_worker : str
        Memory booking for each worker. Can be specified either in megabytes
        (e.g., ``mem_per_worker = 1500MB``) or gigabytes (e.g., ``mem_per_worker = "2GB"``).
        If `mem_per_worker` is `"auto"` it is attempted to infer a sane default value
        from the chosen partition, e.g., for ``partition = "8GBS"`` `mem_per_worker` is
        automatically set to the allowed maximum of `'8GB'`. On the IBM POWER
        partition "E880", `mem_per_worker` is set to 16 GB if not provided.
        Note, even in queues with guaranteed memory bookings, it is possible to allocate less
        memory than the allowed maximum per worker to spawn numerous low-memory
        workers. See Examples for details.
    cores_per_worker : None or int
        Number of CPU cores allocated for each worker. If `None`, core-count
        is set based on partition settings (`DefMemPerCPU`) with respect to
        CPU architecture (minimum 1 on x86_64, and 4 on IBM POWER).
    n_workers_startup : int
        Number of spawned workers to wait for. If `n_workers_startup` is `1` (default),
        the code does not proceed until either 1 SLURM job is running or the
        `timeout` interval has been exceeded.
    timeout : int
        Number of seconds to wait for requested workers to start (see `n_workers_startup`).
    interactive : bool
        If `True`, user input is queried in case not enough workers (set by `n_workers_startup`)
        could be started in the provided waiting period (determined by `timeout`).
        The code waits `interactive_wait` seconds for a user choice - if none is
        provided, it continues with the current number of running workers (if greater
        than zero). If `interactive` is `False` and no worker could not be started
        within `timeout` seconds, a `TimeoutError` is raised.
    interactive_wait : int
        Countdown interval (seconds) to wait for a user response in case fewer than
        `n_workers_startup` workers could be started. If no choice is provided within
        the given time, the code automatically proceeds with the current number of
        active dask workers.
    start_client : bool
        If `True`, a distributed computing client is launched and attached to
        the dask worker cluster. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.
    job_extra : list
        Extra sbatch parameters to pass to SLURMCluster.
    **kwargs : dict
        Additional keyword arguments can be used to control job-submission details.

    Returns
    -------
    proc : object
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise).

    Examples
    --------
    The following command launches 10 SLURM workers with 2 gigabytes memory each
    in the `8GBS` partition

    >>> client = esi_cluster_setup(n_workers=10, partition="8GBS", mem_per_worker="2GB")

    Use default settings to start 2 SLURM workers in the IBM POWER E880 partition
    (allocating 4 cores and 16 GB memory per worker)

    >>> client = esi_cluster_setup(partition="E880")

    The underlying distributed computing cluster can be accessed using

    >>> client.cluster

    Notes
    -----
    The employed parallel computing engine relies on the concurrent processing library
    `Dask <https://docs.dask.org/en/latest/>`_. Thus, the distributed computing
    clients generated here are in fact instances of :class:`distributed.Client`.
    This function specifically acts  as a wrapper for :class:`dask_jobqueue.SLURMCluster`.
    Users familiar with Dask in general and its distributed scheduler and cluster
    objects in particular, may leverage Dask's entire API to fine-tune parallel
    processing jobs to their liking (if wanted).

    See also
    --------
    dask_jobqueue.SLURMCluster : launch a dask cluster of SLURM workers
    slurm_cluster_setup : start a distributed Dask cluster of parallel processing workers using SLURM
    local_cluster_setup : start a local Dask multi-processing cluster on the host machine
    cluster_cleanup : remove dangling parallel processing worker-clusters
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Backwards compatibility: legacy keywords are converted to new nomenclature
    if any(kw in kwargs for kw in __deprecated__):
        log.warning(__deprecation_wrng__)
        n_workers = kwargs.pop("n_jobs", n_workers)                             # type: ignore
        mem_per_worker = kwargs.pop("mem_per_job", mem_per_worker)
        n_workers_startup = kwargs.pop("n_jobs_startup", n_workers_startup)     # type: ignore
        log.debug("Set `n_workers = n_jobs`, `mem_per_worker = mem_per_job`\
                  and `n_workers_startup = n_jobs_startup`")

    # Don't start a new cluster on top of an existing one
    try:
        client = get_client()
        log.debug("Found existing client")
        if count_online_workers(client.cluster) == 0:
            log.debug("No active workers detected in %s", str(client))
            cluster_cleanup(client)
        else:
            log.info("Found existing parallel computing client %s. \
                     Not starting new cluster.", str(client))
            if start_client:
                return client
            return client.cluster
    except ValueError:
        log.debug("No existing clients detected")

    # Check if SLURM's `sinfo` can be accessed
    log.debug("Test if `sinfo` is available")
    proc = subprocess.Popen("sinfo",
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            text=True, shell=True)
    _, err = proc.communicate()

    # Any non-zero return-code means SLURM is not ready to use
    if proc.returncode != 0:

        # SLURM is not installed: either allocate `LocalCluster` or just leave
        if proc.returncode > 0:
            if interactive:
                msg = "{name:s} SLURM does not seem to be installed on this machine " +\
                    "({host:s}). Do you want to start a local multi-processing " +\
                    "computing client instead? "
                startLocal = user_yesno(msg.format(name=funcName, host=socket.gethostname()),
                                        default="no")
            else:
                startLocal = True
            if startLocal:
                return local_cluster_setup(interactive=interactive, start_client=start_client)

        # SLURM is installed, but something's wrong
        msg = "Cannot access SLURM queuing system from node %s: %s "
        log.error(msg, socket.gethostname(), err)
        raise IOError("%s %s"%(funcName, msg%(socket.gethostname(), err)))

    # Use default by-worker process count or extract it from anonymous keyword args (if provided)
    processes_per_worker = kwargs.pop("processes_per_worker", 1)
    log.debug("Found `sinfo`, set `processes_per_worker` to %d", processes_per_worker)

    # Get micro-architecture of submitting host
    mArch = platform.machine()

    # If partition is "auto" use `mem_per_worker` to pick pseudo-optimal partition
    # Note: the `np.where` gymnastic below is necessary since `argmin` refuses
    # to return multiple matches; if `mem_per_worker` is 12, then ``memDiff = [4, 4, ...]``,
    # however, 8GB won't fit a 12GB worker, so we have to pick the second match 16GB
    if isinstance(partition, str):
        if partition == "auto":
            if not isinstance(mem_per_worker, str) or mem_per_worker.find("estimate_memuse:") < 0:
                msg = "Cannot auto-select partition without first invoking memory estimation in `ParallelMap`. "
                log.error(msg)
                raise IOError("%s %s"%(funcName, msg))
            memEstimate = int(mem_per_worker.replace("estimate_memuse:", ""))
            mem_per_worker = "auto"
            log.info("Automatically selecting SLURM partition...")
            availPartitions = _get_slurm_partitions()
            if mArch == "x86_64":
                gbQueues = np.unique([int(queue.split("GB")[0]) for queue in availPartitions if queue[0].isdigit()])
                memDiff = np.abs(gbQueues - memEstimate)
                queueIdx = np.where(memDiff == memDiff.min())[0][-1]
                partition = "{}GBXS".format(gbQueues[queueIdx])
            else:
                partition = "E880"
                mem_per_worker = f"{memEstimate} GB"
            msg = "Picked partition %s based on estimated memory consumption of %d GB"
            log.info(msg, partition, memEstimate)
        else:
            if (partition == "E880" and mArch == "x86_64") or \
               (mArch == "ppc64le" and partition != "E880"):
                otherArch = list(set(["x86_64", "ppc64le"]).difference([mArch]))[0]
                msg = "Cannot start SLURM workers in partition %s with " +\
                    "architecture %s from submitting host with architecture %s. " +\
                    "Start x86_64 workers from esi-svhpc{1,2,3} and POWER workers from the hub."
                raise ValueError(msg%(partition, otherArch, mArch))

    # Convert "auto" memory selection query to `None` for easier handling below`
    if isinstance(mem_per_worker, str) and mem_per_worker == "auto":
        mem_per_worker = None                                       # type: ignore
        log.debug("Using auto-memory selection")

    # If not explicitly provided, extract by-worker CPU core count from
    # partition via `DefMeMPerCPU` and `mem_per_worker`
    if cores_per_worker is None:
        try:
            log.debug("Using `scontrol` to get partition info")
            pc = subprocess.run("scontrol -o show partition {}".format(partition),
                                capture_output=True, check=True, shell=True, text=True)
            defMem = int(pc.stdout.strip().partition("DefMemPerCPU=")[-1].split()[0])
            log.debug("Found DefMemPerCPU=%d", defMem)
        except Exception as exc:
            msg = "Cannot fetch available memory per CPU in SLURM: %s"
            log.error(msg, str(exc))
            raise IOError("%s %s"%(funcName, msg%(str(exc))))

        # Use ESI-specific x86_64 partition layout (8GB(S/X/L), 16GB(S/X/L), ... )
        # to get infer memory and core count; use "sane" (fat quotes)
        # defaults on IBM POWER (if nothing was provided, use 4 cores/16 GB per worker)
        if mArch == "x86_64":
            memPerCore = 8000
        else:
            if mem_per_worker is not None:
                try:
                    memVal = float(mem_per_worker.replace("MB", "").replace("GB", ""))
                except:
                    raise ValueError("Invalid value of `mem_per_worker`: %s"%mem_per_worker)
                if memVal <= 0:
                    raise ValueError("Value of `mem_per_worker` has to be > 0!")
                if "MB" in mem_per_worker:
                    defMem = int(memVal)
                else:
                    defMem = int(round(memVal * 1000))
                log.debug("Found `mem_per_worker` set to %d MB", defMem)
            else:
                defMem = 16000                  # default to 4 cores per worker if no mem req was specified
                mem_per_worker = f"{defMem} MB"
                log.debug("No `mem_per_worker` specified, using default of %d MB", defMem)
            memPerCore = 4000

        # Set core-count per worker (applies to both x86_64 and ppc64le)
        cores_per_worker = max(1, int(defMem / memPerCore))
        log.debug("Derived core-count from partition: `cores_per_worker=%d`", cores_per_worker)

    # Determine if `job_extra`` is a list (this is also checked in `slurm_cluster_setup`,
    # but we may need to append to it, so ensure that's possible)
    if not isinstance(job_extra, list):
        msg = "`job_extra` has to be a list, not %s"
        log.error(msg, str(type(job_extra)))
        raise TypeError("%s %s"%(funcName, msg%(str(type(job_extra)))))

    # If '--output' was not provided, append default output folder to `job_extra`
    if not any(option.startswith("--output") or option.startswith("-o") for option in job_extra):
        log.debug("Auto-populating `--output` setting for sbatch")
        usr = getpass.getuser()
        slurm_wdir = "/cs/slurm/{usr:s}/{usr:s}_{date:s}"
        slurm_wdir = slurm_wdir.format(usr=usr,
                                       date=datetime.now().strftime('%Y%m%d-%H%M%S'))
        os.makedirs(slurm_wdir, exist_ok=True)
        log.debug("Using %s for slurm logs", slurm_wdir)
        out_files = os.path.join(slurm_wdir, "slurm-%j.out")
        job_extra.append("--output={}".format(out_files))
        log.debug("Setting `--output=%s`", out_files)

    # Let the SLURM-specific setup function do the rest (returns client or cluster)
    return slurm_cluster_setup(partition, cores_per_worker, n_workers,      # type: ignore
                               processes_per_worker, mem_per_worker,
                               n_workers_startup, timeout, interactive,
                               interactive_wait, start_client, job_extra,
                               invalid_partitions=["DEV", "ESI"], **kwargs)


# Setup SLURM cluster
def slurm_cluster_setup(
        partition: str = "partition_name",
        n_cores: int = 1,
        n_workers: int = 1,
        processes_per_worker: int = 1,
        mem_per_worker: str = "1GB",
        n_workers_startup: int = 1,
        timeout: int = 60,
        interactive: bool = True,
        interactive_wait: int = 10,
        start_client: bool = True,
        job_extra: List = [],
        invalid_partitions: List = [],
        **kwargs: Optional[Any]) -> Union[Client, SLURMCluster, None]:
    """
    Start a distributed Dask cluster of parallel processing workers using SLURM

    **NOTE** If you are working on the ESI HPC cluster, please use
    :func:`~acme.esi_cluster_setup` instead!

    Parameters
    ----------
    partition : str
        Name of SLURM partition/queue to use
    n_cores : int
        Number of CPU cores per SLURM worker
    n_workers : int
        Number of SLURM workers (=jobs) to spawn
    processes_per_worker : int
        Number of processes to use per SLURM job (=worker). Should be greater
        than one only if the chosen partition contains nodes that expose multiple
        cores per job.
    mem_per_worker : str
        Memory allocation for each worker
    n_workers_startup : int
        Number of spawned SLURM workers to wait for. The code does not return until either
        `n_workers_startup` SLURM jobs are running or the `timeout` interval (see
        below) has been exceeded.
    timeout : int
        Number of seconds to wait for requested workers to start (see `n_workers_startup`).
    interactive : bool
        If `True`, user input is queried in case not enough workers (set by `n_workers_startup`)
        could be started in the provided waiting period (determined by `timeout`).
        The code waits `interactive_wait` seconds for a user choice - if none is
        provided, it continues with the current number of running workers (if greater
        than zero). If `interactive` is `False` and no worker could not be started
        within `timeout` seconds, a `TimeoutError` is raised.
    interactive_wait : int
        Countdown interval (seconds) to wait for a user response in case fewer than
        `n_workers_startup` workers could be started. If no choice is provided within
        the given time, the code automatically proceeds with the current number of
        active dask workers.
    start_client : bool
        If `True`, a distributed computing client is launched and attached to
        the dask worker cluster. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.
    job_extra : list
        Extra sbatch parameters to pass to SLURMCluster.
    invalid_partition : list
        List of partition names (strings) that are not available for launching
        dask workers.

    Returns
    -------
    proc : object or None
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise). If no SLURM workers
        can be started within the given timeout interval, `proc` is set
        to `None`.

    See also
    --------
    dask_jobqueue.SLURMCluster : launch a dask cluster of SLURM workers
    esi_cluster_setup : start a SLURM worker cluster on the ESI HPC infrastructure
    local_cluster_setup : start a local Dask multi-processing cluster on the host machine
    cluster_cleanup : remove dangling parallel processing worker-clusters
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Backwards compatibility: legacy keywords are converted to new nomenclature
    if any(kw in kwargs for kw in __deprecated__):
        log.warning(__deprecation_wrng__)
        n_workers = kwargs.pop("n_jobs", n_workers)                                 # type: ignore
        processes_per_worker = kwargs.pop("workers_per_job", processes_per_worker)  # type: ignore
        mem_per_worker = kwargs.pop("mem_per_job", mem_per_worker)                  # type: ignore
        n_workers_startup = kwargs.pop("n_jobs_startup", n_workers_startup)         # type: ignore
        log.debug("Set `n_workers = n_jobs`, `processes_per_worker = workers_per_job`, \
                  `mem_per_worker = mem_per_job` \
                  and `n_workers_startup = n_jobs_startup`")

    # Retrieve all partitions currently available in SLURM
    availPartitions = _get_slurm_partitions()

    # Make sure we're in a valid partition
    if partition not in availPartitions:
        valid = list(set(availPartitions).difference(invalid_partitions))
        lgl = "'" + "or '".join(opt + "' " for opt in valid)
        msg = "Invalid partition selection %s, available SLURM partitions are %s"
        log.error(msg, str(partition), lgl)
        raise ValueError("%s %s"%(funcName, msg%(str(partition), lgl)))
    log.debug("Found `partition = %s`", partition)

    # Parse worker count
    try:
        scalar_parser(n_workers, varname="n_workers", ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        log.error("Error parsing `n_workers`")
        raise exc
    log.debug("Using `n_workers = %d`", n_workers)

    # Get requested memory per worker
    if isinstance(mem_per_worker, str):
        if mem_per_worker == "auto":
            mem_per_worker = None                                       # type: ignore
            log.debug("Using auto-memory selection")
    if mem_per_worker is not None:
        msg = "`mem_per_worker` has to be a valid memory specifier (e.g., '8GB', '12000MB'), not %s"
        if not isinstance(mem_per_worker, str):
            log.error(msg, str(type(mem_per_worker)))
            raise TypeError("%s %s"%(funcName, msg%(str(type(mem_per_worker)))))
        if not any(szstr in mem_per_worker for szstr in ["MB", "GB"]):
            log.error(msg, mem_per_worker)
            raise ValueError("%s %s"%(funcName, msg%(mem_per_worker)))
        memNumeric = mem_per_worker.replace("MB", "").replace("GB", "")
        log.debug("Found `mem_per_worker = %s` in input args", mem_per_worker)
        try:
            memVal = float(memNumeric)
        except:
            log.error(msg, mem_per_worker)
            raise ValueError("%s %s"%(funcName, msg%(mem_per_worker)))
        if memVal <= 0:
            log.error(msg, mem_per_worker)
            raise ValueError("%s %s"%(funcName, msg%(mem_per_worker)))

    # Parse worker-waiter count
    try:
        scalar_parser(n_workers_startup, varname="n_workers_startup", ntype="int_like", lims=[0, np.inf])
    except Exception as exc:
        log.error("Error parsing `n_workers_startup`")
        raise exc
    log.debug("Using `n_workers_startup = %d`", n_workers_startup)

    # Get memory limit (*in MB*) of chosen partition (guaranteed to exist, cf. above)
    log.debug("Use `scontrol` to fetch partition's memory limit")
    pc = subprocess.run("scontrol -o show partition {}".format(partition),
                        capture_output=True, check=True, shell=True, text=True)
    try:
        mem_lim = int(pc.stdout.strip().partition("MaxMemPerCPU=")[-1].split()[0])
    except IndexError:
        mem_lim = np.inf                                            # type: ignore
    log.debug("Found a limit of  %s MB", str(mem_lim))

    # Consolidate requested memory with chosen partition (or assign default memory)
    if mem_per_worker is None:
        if np.isinf(mem_lim):
            try:
                mem_per_worker = pc.stdout.strip().partition("DefMemPerCPU=")[-1].split()[0] + "MB"
            except IndexError:
                raise ValueError("Cannot infer any default memory setting from partition %s"%partition)
        else:
            mem_per_worker = str(mem_lim) + "MB"
        log.debug("Using partition limit of %s MB", str(mem_lim))
    else:
        if "MB" in mem_per_worker:
            mem_req = int(memVal)
        else:
            mem_req = int(round(memVal * 1000))
        if mem_req > mem_lim:
            msg = "`mem_per_worker` exceeds limit of %d MB for partition %s. " +\
                "Capping memory at partition limit. "
            log.warning(msg, mem_lim, partition)
            mem_per_worker = str(int(mem_lim)) + "MB"

    # Parse requested timeout period
    try:
        scalar_parser(timeout, varname="timeout", ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        log.error("Error parsing `timeout`")
        raise exc
    log.debug("Using `timeout = %d`", timeout)

    # Parse requested interactive waiting period
    try:
        scalar_parser(interactive_wait, varname="interactive_wait", ntype="int_like",
                      lims=[0, np.inf])
    except Exception as exc:
        log.error("Error parsing `interactive_wait`")
        raise exc
    log.debug("Using `interactive_wait = %d`", interactive_wait)

    # Determine if cluster allocation is happening interactively
    if not isinstance(interactive, bool):
        msg = "`interactive` has to be Boolean, not %s"
        log.error(msg, str(type(interactive)))
        raise TypeError("%s %s"%(funcName, msg%(str(type(interactive)))))
    log.debug("Using `interactive = %s`", str(interactive))

    # Determine if a dask client was requested
    if not isinstance(start_client, bool):
        msg = "`start_client` has to be Boolean, not %s"
        log.error(msg, str(type(start_client)))
        raise TypeError("%s %s"%(funcName, msg%(str(type(start_client)))))
    log.debug("Using `start_client = %s`", str(start_client))

    # Determine if job_extra is a list
    if not isinstance(job_extra, list):
        msg = "`job_extra` has to be List, not %s"
        log.error(msg, str(type(job_extra)))
        raise TypeError("%s %s"%(funcName, msg%(str(type(job_extra)))))

    # Determine if job_extra options are valid
    for option in job_extra:
        msg = "`job_extra` has to be a valid sbatch option, not %s"
        if not isinstance(option, str):
            log.error(msg, str(type(option)))
            raise TypeError("%s %s"%(funcName, msg%(str(type(option)))))
        if not option[0] == "-":
            msg = "`job_extra` options must be flagged with - or --"
            log.error(msg)
            raise ValueError("%s %s"%(funcName, msg))
    log.debug("Using `job_extra = %s`", str(job_extra))

    # Ensure validity of requested worker processes
    try:
        scalar_parser(processes_per_worker, varname="processes_per_worker",
                      ntype="int_like", lims=[1, 16])
    except Exception as exc:
        log.error("Error parsing `processes_per_worker`")
        raise exc
    log.debug("Using `processes_per_worker = %d`", processes_per_worker)

    # Check for sanity of requested core count
    try:
        scalar_parser(n_cores, varname="n_cores",
                      ntype="int_like", lims=[1, np.inf])
    except Exception as exc:
        log.error("Error parsing `n_cores`")
        raise exc
    log.debug("Using `n_cores = %d`", n_cores)

    # Check validity of '--output' option if provided
    userOutSpec = [option.startswith("--output") or option.startswith("-o") for option in job_extra]
    if any(userOutSpec):
        userOut = job_extra[userOutSpec.index(True)]
        outSpec = userOut.split("=")
        if len(outSpec) != 2:
            msg = "SLURM output directory must be specified using -o/--output=/path/to/file, not %s"
            log.error(msg, userOut)
            raise ValueError("%s %s"%(funcName, msg%(userOut)))
        slurm_wdir = os.path.split(outSpec[1])[0]
        if len(slurm_wdir) > 0 and not os.path.isdir(os.path.expanduser(slurm_wdir)):
            msg = "SLURM output location has to be an existing directory, not %s"
            log.error(msg, slurm_wdir)
            raise ValueError("%s %s"%(funcName, msg%(slurm_wdir)))
    else:
        slurm_wdir = None
    log.debug("Using `local_directory = %s`", slurm_wdir)

    # Create `SLURMCluster` object using provided parameters
    log.debug("Instantiating `SLURMCluster` object")
    cluster = SLURMCluster(cores=n_cores,
                           memory=mem_per_worker,
                           processes=processes_per_worker,
                           local_directory=slurm_wdir,
                           queue=partition,
                           python=sys.executable,
                           job_directives_skip=["-t", "--mem"],
                           job_extra_directives=job_extra)
                           # interface="asdf", # interface is set via `psutil.net_if_addrs()`

    # Compute total no. of workers and up-scale cluster accordingly
    if n_workers_startup < n_workers:
        msg = "Requested worker-count %d exceeds `n_workers_startup = %d`, " +\
            "waiting for %d workers to come online"
        log.debug(msg, n_workers, n_workers_startup, n_workers_startup)
    cluster.scale(n_workers)

    # Fire up waiting routine to avoid returning an undercooked cluster
    if _cluster_waiter(cluster, funcName, n_workers, timeout, interactive, interactive_wait):
        return None

    # Kill a zombie cluster in non-interactive mode
    if not interactive and count_online_workers(cluster) == 0:
        cluster_cleanup(Client(cluster))
        msg = "SLURM workers could not be started within given time-out " +\
              "interval of %d seconds"
        log.error(msg, timeout)
        raise TimeoutError("%s %s"%(funcName, msg%(timeout)))

    # Highlight how to connect to dask performance monitor
    msg = "Parallel computing client ready, dashboard accessible at %s"
    log.info(msg, cluster.dashboard_link)

    # If client was requested, return that instead of the created cluster
    if start_client:
        return Client(cluster)
    return cluster


def _get_slurm_partitions() -> List:
    """
    Local helper to fetch all partitions defined in SLURM
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Retrieve all partitions currently available in SLURM
    log.debug("Use `sinfo` to fetch available partitions")
    proc = subprocess.Popen("sinfo -h -o %P",
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            text=True, shell=True)
    out, err = proc.communicate()

    # Any non-zero return-code means SLURM is not ready to use
    if proc.returncode != 0:
        msg = "Error fetching SLURM partition setup from node %s: %s"
        log.error(msg, socket.gethostname(), err)
        raise IOError("%s %s"%(funcName, msg%(socket.gethostname(), err)))

    # Return formatted subprocess shell output
    log.debug("Found partitions: %s", out)
    return out.split()


def _cluster_waiter(
        cluster: SLURMCluster,
        funcName: str,
        total_workers: int,
        timeout: int,
        interactive: bool,
        interactive_wait: int) -> bool:
    """
    Local helper that can be called recursively
    """

    # Wait until all workers have been started successfully or we run out of time
    wrkrs = count_online_workers(cluster)
    to = str(timedelta(seconds=timeout))[2:]
    fmt = "{desc}: {n}/{total} \t[elapsed time {elapsed} | timeout at " + to + "]"
    ani = tqdm(desc="{} SLURM workers ready".format(funcName), total=total_workers,
               leave=True, bar_format=fmt, initial=wrkrs, position=0)
    counter = 0
    while count_online_workers(cluster) < total_workers and counter < timeout:
        time.sleep(1)
        counter += 1
        ani.update(max(0, count_online_workers(cluster) - wrkrs))
        wrkrs = count_online_workers(cluster)
        ani.refresh()   # force refresh to display elapsed time every second
    ani.close()

    # If we ran out of time before all workers could be started, ask what to do
    if counter == timeout and interactive:                              # pragma: no cover
        msg = "SLURM workers could not be started within given time-out " +\
              "interval of %d seconds"
        log.info(msg, timeout)
        query = "{name:s} Do you want to [k]eep waiting for 60s, [a]bort or " +\
                "[c]ontinue with {wrk:d} workers?"
        choice = user_input(query.format(name=funcName, wrk=wrkrs),
                            valid=["k", "a", "c"], default="c", timeout=interactive_wait)

        if choice == "k":
            return _cluster_waiter(cluster, funcName, total_workers, 60, True, 60)
        elif choice == "a":
            log.info("Closing cluster...")
            cluster_cleanup(Client(cluster))
            return True
        else:
            if wrkrs == 0:
                query = "{} Cannot continue with 0 workers. Do you want to " +\
                        "[k]eep waiting for 60s or [a]bort?"
                choice = user_input(query.format(funcName), valid=["k", "a"],
                                    default="a", timeout=60)
                if choice == "k":
                    _cluster_waiter(cluster, funcName, total_workers, 60, True, 60)
                else:
                    log.info("Closing cluster...")
                    cluster_cleanup(Client(cluster))
                    return True

    return False


def local_cluster_setup(
        n_workers: Optional[int] = None,
        mem_per_worker: Optional[str] = None,
        interactive: bool = True,
        start_client: bool = True) -> Union[Client, LocalCluster, None]:
    """
    Start a local distributed Dask multi-processing cluster

    Parameters
    ----------
    n_workers : int
        Number of local workers to start (this should align with the locally
        available hardware, see :class:`distributed.LocalCluster` for details)
    mem_per_worker : str
        Memory cap for each local worker (corresponds to the `memory_limit`
        keyword of a :class:`distributed.worker.Worker`)
    interactive : bool
        If `True`, a confirmation dialog is displayed to ensure proper encapsulation
        of calls to `local_cluster_setup` inside a script's main module block.
        See Notes for details. If `interactive` is `False`, the dialog is not shown.
    start_client : bool
        DEPRECATED. Will be removed in next release.
        If `True`, a distributed computing client is launched and attached to
        the workers. If `start_client` is `False`, only a distributed
        computing cluster is started to which compute-clients can connect.

    Returns
    -------
    proc : object
        A distributed computing client (if ``start_client = True``) or
        a distributed computing cluster (otherwise). If a client cannot
        be started, `proc` is set to `None`.

    Notes
    -----
    The way Python spawns new processes requires an explicit separation of initialization
    code (i.e., code blocks that should only be executed once) from the actual
    program code. Specifically, everything that is supposed to be invoked only
    once by the parent spawner must be encapsulated in a script's main block.
    Otherwise, initialization code is not only run once by the parent process but
    executed by every child process at import time.

    This means that starting a local multi-processing cluster *has* to be wrapped
    inside a script's main module block, otherwise, every child process created by
    the multi-processing cluster starts a multi-processing cluster itself and so
    on escalating to an infinite recursion. Thus, if `local_cluster_setup` is
    called inside a script, it has to be encapsulated in the script's main module
    block, i.e.,

    .. code-block:: python

        if __name__ == "__main__":
            ...
            local_cluster_setup()
            ...

    Note that this capsulation is **only** required inside Python scripts. Launching
    `local_cluster_setup` in the (i)Python shell or inside a Jupyter notebook
    does not suffer from this problem.
    A more in-depth technical discussion of this limitation can be found in
    `Dask's GitHub issue tracker <https://github.com/dask/distributed/issues/2520>`_.

    Examples
    --------
    The following command launches a local distributed computing cluster using
    all CPU cores available on the host machine

    >>> client = local_cluster_setup()

    The underlying distributed computing cluster can be accessed using

    >>> client.cluster

    See also
    --------
    distributed.LocalCluster : create local worker cluster
    esi_cluster_setup : Start a distributed Dask cluster using SLURM
    cluster_cleanup : remove dangling parallel processing worker-clusters
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Determine if cluster allocation is happening interactively
    if not isinstance(interactive, bool):
        msg = "`interactive` has to be Boolean, not %s"
        log.error(msg, str(type(interactive)))
        raise TypeError("%s %s"%(funcName, msg%(str(type(interactive)))))
    log.debug("Using `interactive = %s`", str(interactive))

    # Determine if a dask client was requested
    if not isinstance(start_client, bool) or start_client is False:
        log.warning("The keyword `start_client` in `local_cluster_setup` is DEPRECATED and will be ignored.")
        start_client = True

    # ...if not, print warning/info message
    if not is_jupyter():
        msg = """\
        If you use a script to start a local parallel computing client, please ensure
        the call to `local_cluster_setup` is wrapped inside a main module block, i.e.,

            if __name__ == "__main__":
                ...
                local_cluster_setup()
                ...

        Otherwise, a RuntimeError is raised due to an infinite recursion triggered by
        new processes being started before the calling process can finish its bootstrapping
        phase.
        """
        msg = textwrap.dedent(msg)
        log.debug(msg)

    # Additional safe-guard: if a script is executed, double-check with the user
    # for proper main idiom usage
    if interactive:                                                     # pragma: no cover
        msg = "{name:s} If launched from a script, did you wrap your code " +\
            "inside a __main__ module block?"
        if not user_yesno(msg.format(name=funcName), default="no"):
            return None

    # Start the actual distributed client
    if n_workers is not None or mem_per_worker is not None:
        msg = "Starting `LocalCluster` with `n_workers = %s` and `memory_limit = %s`"
        log.debug(msg, str(n_workers), str(mem_per_worker))
        cluster = LocalCluster(n_workers=n_workers, memory_limit=mem_per_worker)
        client = Client(cluster)
    else:
        client = Client()
    msg = "Local parallel computing client ready, dashboard accessible at %s"
    log.info(msg, client.cluster.dashboard_link)
    if start_client:
        return client
    return client.cluster


def cluster_cleanup(client: Optional[Client] = None) -> None:
    """
    Stop and close dangling parallel processing workers

    Parameters
    ----------
    client : dask distributed computing client or None
        Either a  :class:`distributed.Client` or `None`. If `None`, a
        global client is queried for and shut-down if found (without confirmation!).

    Returns
    -------
    Nothing : None

    See also
    --------
    esi_cluster_setup : Launch SLURM workers on the ESI compute cluster
    slurm_cluster_setup : start a distributed Dask cluster of parallel processing workers using SLURM
    local_cluster_setup : start a local Dask multi-processing cluster on the host machine
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Attempt to establish connection to dask client
    if client is None:
        try:
            client = get_client()
        except ValueError:
            log.warning("No dangling clients or clusters found.")
            return
        except Exception as exc:                                        # pragma: no cover
            log.error("Error looking for dask client")
            raise exc
    else:
        if not isinstance(client, Client):
            msg = "`client` has to be a dask client object, not %s"
            log.error(msg, str(type(client)))
            raise TypeError("%s %s"%(funcName, msg%(str(type(client)))))
    log.debug("Found client %s", str(client))

    # Prepare message for prompt
    if client.cluster.__class__.__name__ == "LocalCluster":
        userClust = "LocalCluster hosted on {}".format(client.scheduler_info()["address"])
    else:
        userName = getpass.getuser()
        outDir = client.cluster.job_header.partition("--output=")[-1]
        jobID = outDir.partition("{}_".format(userName))[-1].split(os.sep)[0]
        userClust = "cluster {0}_{1}".format(userName, jobID)
    nWorkers = count_online_workers(client.cluster)

    # First gracefully shut down all workers, then close client
    client.retire_workers(list(client.scheduler_info()['workers']), close_workers=True)
    client.close()
    try:
        client.cluster.close()
    except Exception as exc:
        log.warning("Could not gracefully shut down cluster: %s", str(exc))

    # Communicate what just happened and get outta here
    msg = "Successfully shut down %s containing %d workers"
    log.info(msg, userClust, nWorkers)

    return


def count_online_workers(cluster: SLURMCluster) -> int:
    """
    Local replacement for the late `._count_active_workers` class method
    """
    return len([w["memory_limit"] for w in cluster.scheduler_info["workers"].values()])



#
# Interface for integrating with Syncopy
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import sys
import logging

# Differentiate b/w being imported as Syncopy sub-package or standalone
# ACME module: do not attempt to import syncopy but instead see if it has
# already been imported; then ACME should use Syncopy's built-in logging
if "syncopy" in sys.modules:                                            # pragma: no cover
    import syncopy as spy                                               # type: ignore
    log = logging.getLogger("syncopy")
    scalar_parser = lambda var, varname="", ntype=None, lims=None : \
        spy.shared.parsers.scalar_parser(var, varname=varname, ntype=ntype, lims=lims)
else:
    from .shared import _scalar_parser as scalar_parser
    log = logging.getLogger("ACME")



#
# ACME's logging facilities
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import warnings
import inspect
import logging
import os
from logging import handlers
from typing import Optional, Any, List

__all__: List[str] = []


def prepare_log(
        logname: str = "ACME",
        logfile: Optional[str] = None,
        verbose: Optional[bool] = None) -> None:
    """
    Convenience function to set up ACME logger

    Parameters
    ----------
    logname : str
        Name of the logger to set up
    logfile : None or str
        If `None`, logging information is streamed to stdout only, otherwise
        `logfile` is interpreted as path to a file used for logging.
    verbose : bool or None
        If `None`, the logging-level only contains messages of `'INFO'` priority and
        higher (`'WARNING'` and `'ERROR'`). If `verbose` is `True`, logging is
        performed on ``DEBUG`', `'INFO`', `'WARNING'` and `'ERROR'` levels. If
        `verbose` is `False` only `'WARNING'` and `'ERROR'` messages are propagated.

    Returns
    -------
    log : logger object
        A Python :class:`logging.Logger` instance
    """

    # For later reference: dynamically fetch name of current function
    funcName = "<{}>".format(inspect.currentframe().f_code.co_name)     # type: ignore

    # Ensure `logname` can be processed
    if not isinstance(logname, str):
        msg = "%s `caller` has to be a string, not %s"
        raise TypeError(msg%(funcName, str(type(logname))))

    # Basal sanity check for Boolean flag
    if verbose is not None and not isinstance(verbose, bool):
        msg = "%s `verbose` has to be `True`, `False` or `None`, not %s"
        raise TypeError(msg%(funcName, str(type(verbose))))

    # Either parse provided `logfile` or set up an auto-generated file
    if logfile is not None and os.path.isfile(logfile):
        msg = "%s log-file %s already exists, appending to it"
        warnings.warn(msg%(logname, logfile))

    # Add custom "announce" level to logger
    announceLvl = logging.INFO + 5
    logging.addLevelName(announceLvl, "ANNOUNCE")
    logging.ANNOUNCE = announceLvl                                      # type: ignore

    def announce(
            self,
            msg: str,
            *args: Any,
            **kwargs: Optional[Any]) -> None:
        """
        Log 'msg % args' with severity 'ANNOUNCE'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.announce("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        if self.isEnabledFor(announceLvl):
            self._log(announceLvl, msg, args, **kwargs)

    logging.getLoggerClass().announce = announce                        # type: ignore
    logging.announce = announce                                         # type: ignore

    # Fetch/set up ACME logger
    log = logging.getLogger(logname)

    # Set logging verbosity based on `verbose` flag
    if verbose is None:
        loglevel = logging.INFO
    elif verbose is True:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.WARNING
    log.setLevel(loglevel)

    # Create logging formatters
    acmeFmt = "%(name)s %(levelname)s <%(funcName)s> %(message)s"
    streamFrmt = AcmeFormatter(acmeFmt, color=True)
    fileFrmt = AcmeFormatter(acmeFmt, color=False)

    # Upon package import, create stdout handler + memory handler for
    # temporary buffering of all log messages
    if len(log.handlers) == 0:
        stdoutHandler = logging.StreamHandler()
        log.addHandler(stdoutHandler)
        memHandler = handlers.MemoryHandler(1000,
                                            flushLevel=logging.ERROR,
                                            target=None,
                                            flushOnClose=True)
        log.addHandler(memHandler)

    # If log-file creation was requested, add a target to the initially
    # created `MemoryHandler` (initiated by `ACMEdaemon`)
    if logfile is not None:
        memHandler = [h for h in log.handlers if isinstance(h, handlers.MemoryHandler)][0]
        fileHandler = logging.FileHandler(logfile)
        fileHandler.setLevel(loglevel)
        fileHandler.setFormatter(fileFrmt)
        memHandler.setTarget(fileHandler)

    # No matter if called for the first or n-th time, (re)set log-level
    # and formatter
    for h in log.handlers:
        h.setLevel(loglevel)
        h.setFormatter(streamFrmt)

    return


class AcmeFormatter(logging.Formatter):
    """
    Adapted from https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging/
    """

    def __init__(
            self,
            fmt: str,
            color: bool = True):

        super().__init__()

        if color:
            green = "\x1b[92m"
            gray = "\x1b[90m"
            blue = "\x1b[38;5;39m"
            magenta = "\x1b[35m"
            red = "\x1b[38;5;196m"
            bold = "\x1b[1m"
            reset = "\x1b[0m"
        else:
            green = ""
            gray = ""
            blue = ""
            magenta = ""
            red = ""
            bold = ""
            reset = ""

        fmtName = fmt.partition("%(name)s")
        fmtName = fmtName[0] + bold + fmtName[1] + reset + fmtName[2]   # type: ignore
        fmt = "".join(fmtName)

        fmtLvl = fmt.partition("%(levelname)s")
        fmtDebug = fmtLvl[0] + bold + green + \
            "# " + fmtLvl[1] + " #" + reset + gray + fmtLvl[2] + reset
        fmtInfo = fmtLvl[0] + bold + blue + \
            "- " + fmtLvl[1] + " -" + reset + fmtLvl[2]
        fmtAnnounce = fmtLvl[0] + bold + blue + \
            "> " + fmtLvl[1] + " <" + reset + bold + fmtLvl[2] + reset
        fmtWarn = fmtLvl[0] + bold + magenta + \
            "! " + fmtLvl[1] + " !" + reset + fmtLvl[2]
        fmtError = fmtLvl[0] + bold + red + \
            "| " + fmtLvl[1] + " |" + reset + red + fmtLvl[2] + reset

        fmtAnnounce = "".join(fmtAnnounce).replace("<%(funcName)s>", "")
        fmtInfo = "".join(fmtInfo).replace("<%(funcName)s>", "")

        self.FORMATS = {
            logging.DEBUG: "".join(fmtDebug),
            logging.INFO: "".join(fmtInfo),
            logging.ANNOUNCE: "".join(fmtAnnounce),                     # type: ignore
            logging.WARNING: "".join(fmtWarn),
            logging.ERROR: "".join(fmtError),
            logging.CRITICAL: "".join(fmtError),
        }

    def format(
            self,
            record: logging.LogRecord):
        logFmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(logFmt)
        return formatter.format(record)



#
# User-exposed interface of acme
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import inspect
import numpy as np
import logging
from typing import Any, Callable, Optional, Union, List

# Local imports
from acme import __deprecated__, __deprecation_wrng__, __version__
from .backend import ACMEdaemon
from .logger import prepare_log
from .shared import _scalar_parser, sizeOf
from . import shared as acs

__all__: List["str"] = ["ParallelMap"]

# Fetch logger
log = logging.getLogger("ACME")


# Main context manager for parallel execution of user-defined functions
class ParallelMap(object):

    objName = "<ParallelMap>"
    argv = None
    kwargv = None
    func = None
    n_inputs = None
    _maxArgSize = 1024

    def __init__(
        self,
        func: Callable,
        *args: Any,
        n_inputs: Union[int, str] = "auto",
        write_worker_results: bool = True,
        output_dir: Optional[str] = None,
        result_shape: Optional[tuple[Optional[int], ...]] = None,
        result_dtype: str = "float",
        single_file: bool = False,
        write_pickle: bool = False,
        partition: str = "auto",
        n_workers: Union[int, str] = "auto",
        mem_per_worker: str = "auto",
        setup_timeout: int = 60,
        setup_interactive: bool = True,
        stop_client: Union[bool, str] = "auto",
        verbose: Optional[bool] = None,
        dryrun: bool = False,
        logfile: Optional[Union[bool, str]] = None,
        **kwargs: Optional[Any]) -> None:
        """
        Context manager that executes user-defined functions in parallel

        Parameters
        ----------
        func : callable
            User-defined function to be executed concurrently. Input arguments
            and return values should be "simple" (i.e., regular Python objects or
            NumPy arrays). See Examples and [1]_ for more information.
        args : arguments
            Positional arguments of `func`. Should be regular Python objects
            (lists, tuples, scalars, strings etc.) or NumPy arrays. See
            Examples and [1]_ for more information.
        kwargs : keyword arguments
            Keyword arguments of `func` (if any). Should be regular Python objects
            (lists, tuples, scalars, strings etc.) or NumPy arrays. See Examples
            and [1]_ for more information.
        n_inputs : int or "auto"
            Number of times `func` is supposed to be called in parallel. Usually,
            `n_inputs` does not have to be provided explicitly. If `n_inputs` is
            `"auto"` (default) this quantity is inferred from provided `args` and
            `kwargs`. This estimation may fail due to ambiguous input arguments
            (e.g., `args` and/or `kwargs` contain lists of differing lengths)
            triggering a `ValueError`. Only then is it required to set `n_input`
            manually. See Examples and [1]_ for more information.
        write_worker_results : bool
            If `True`, the return value(s) of `func` is/are saved on disk.
            If `False`, the output of all parallel calls of `func` is collected
            in memory. See Examples as well as [1]_ and [2]_ for more information.
        output_dir : str or None
            Only relevant if `write_worker_results` is `True`. If `output_dir` is `None`
            (default) and `write_worker_results` is `True`, all files auto-generated
            by `ParallelMap` are stored in a directory `'ACME_YYYYMMDD-hhmmss-ffffff'`
            (encoding the current time as YearMonthDay-HourMinuteSecond-Microsecond).
            The path to a custom output directory can be specified via providing
            `output_dir`. See Examples and [1]_ for more information.
        result_shape : tuple or None
            Only relevant if `write_pickle` is `False`. If provided, return
            values of `func` are slotted into a (virtual) dataset (if
            `write_worker_results` is True) or array (otherwise) of shape
            `result_shape`, where a single `None` entry designates the stacking
            dimension. For instance, ``result_shape = (None, 100)`` implies
            that `func` returns a 100-element array which is to be stacked
            along the first dimension for each concurrent call of `func`
            resulting in a ``(n_inputs, 100)`` dataset or array. See Notes
            and Examples for details. See Examples as well as [1]_ and [2]_
            for more information.
        result_dtype : str
            Only relevant if `result_shape` is not `None`. Determines
            the numerical datatype of the dataset laid out by `result_shape`.
            By default, results are stored in `float64` format. See [2]_ for
            more details.
        single_file : bool
            Only relevant if `write_worker_results` is `True` and `write_pickle`
            is `False`. If `single_file` is `False` (default), the results of each parallel
            call of `func` are stored in dedicated HDF5 files, such that the auto-
            generated HDF5 results-container is a collection of symbolic links
            pointing to these files.
            Conversely, if `single_file` is `True`, all parallel workers
            write to the same results container (using a distributed file-locking
            mechanism). See [2]_ for more details.
        write_pickle : bool
            Only relevant if `write_worker_results` is `True`. If `True`,
            the return value(s) of `func` is/are pickled to disk (one
            `'.pickle'`-file per parallel worker). See Examples as well as
            [1]_ and [2]_ for more information.
        partition : str
            Name of SLURM partition to use. If `"auto"` (default), the memory footprint
            of `func` is estimated using dry-run stubs based on randomly sampling
            provided `args` and `kwargs`. Estimated memory usage dictates queue
            auto-selection under the assumption of short run-times (**currently only
            supported on the ESI HPC cluster**). For instance, on a predicted memory
            footprint of 6 GB causes the `"8GBXS"` partition to be selected (minimal
            but sufficient memory and shortest runtime).
            To override auto-selection, provide name of SLURM queue explicitly. See, e.g.,
            :func:`~acme.esi_cluster_setup` for details.
        n_workers : int or "auto"
            Number of SLURM workers (=jobs) to spawn. If `"auto"` (default), then
            ``n_workers = n_inputs``, i.e., every SLURM worker performs a single
            call of `func`.
            If `n_inputs` is large and executing `func` is fast, setting
            ``n_workers = int(n_inputs / 2)`` might be beneficial. See Examples
            as well as [1]_ and [2]_ for more information.
        mem_per_worker : str
            Memory booking for each SLURM worker. If `"auto"` (default), the standard
            value is inferred from the used partition (if possible). See
            :func:`~acme.slurm_cluster_setup` for details.
        setup_timeout : int
            Timeout period (in seconds) for SLURM workers to come online. Refer to
            keyword `timeout` in :func:`~acme.slurm_cluster_setup` for details.
        setup_interactive : bool
            If `True` (default), user input is queried in case not enough SLURM
            workers could be started within `setup_timeout` seconds. If no input
            is provided, the current number of spawned workers is used (even if
            smaller than the amount requested by `n_workers`). If `False`, no user
            choice is requested. Refer to keyword `interactive` in :func:`~acme.slurm_cluster_setup`
        stop_client : bool or "auto"
            If `"auto"` (default), automatically started distributed computing clients
            are shut down at the end of computation, while user-provided clients
            are left untouched. If `False`, automatically started clients are
            left running after completion, user-provided clients are left untouched.
            If `True`, auto-generated clients *and* user-provided clients are
            shut down at the end of the computation. See Examples as well
            as [1]_ and [2]_ for more information.
        verbose : None or bool
            If `None` (default), general run-time information as well as warnings
            and errors are shown. If `True`, additionally debug information is
            shown. If `False`, only warnings and errors are propagated.
            See [2]_ for more details.
        dryrun : bool
            If `True` the user-provided function `func` is executed once using
            one of the input argument tuples prepared for the parallel workers (picked
            at random). If `setup_interactive` is `True`, a prompt asks if the
            actual parallel execution of `func` is supposed to be launched after the
            dry-run. The `dryrun` keyword is intended to to estimate memory consumption
            as well as runtime of worker jobs prior to the actual concurrent
            computation. See [1]_ and [2]_ for more information.
        logfile : None or bool or str
            If `None` (default) and ``write_worker_results = True``,
            a logfile is created alongside the auto-generated on-disk results.
            If `None` and ``write_worker_results = False``, no logfile is
            created. To override this mechanism, either explicitly set
            `logfile` to `True` or `False` to enforce or suppress logfile
            creation.
            Alternatively, the name of a custom log-file can be provided.
            The verbosity of recorded runtime information can be controlled
            via setting `verbose`. See [2]_ for more details.

        Returns
        -------
        results : list
            If `write_worker_results` is `True`, `results` is a list of HDF5 file-names
            containing computed results. If `write_worker_results` is `False`,
            results is a list comprising the actual return values of `func`.

        Examples
        --------
        Call `f` with four different values of `x` while setting `y` to 4:

        .. code-block:: python

            from acme import ParallelMap

            with ParallelMap(f, [2, 4, 6, 8], 4) as pmap:
                results = pmap.compute()

        Collect results in memory (can be slow due to network traffic and may cause
        memory overflow in parent caller):

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], 4, write_worker_results=False) as pmap:
                results = pmap.compute()

        Manually set `n_inputs` in case of argument distribution cannot be determined
        automatically:

        .. code-block:: python

            with ParallelMap(f, [2, 4, 6, 8], y, n_inputs=4, write_worker_results=False) as pmap:
                results = pmap.compute()

        More examples and tutorials are available in the
        `ACME online documentation <https://esi-acme.readthedocs.io>`_.

        Notes
        -----
        Please consult [1]_ for detailed usage information.

        See also
        --------
        esi_cluster_setup : spawn custom SLURM worker clients on the ESI HPC cluster
        local_cluster_setup : start a local Dask multi-processing cluster on the host machine
        ACMEdaemon : Manager class performing the actual concurrent processing

        References
        ----------
        .. [1] https://esi-acme.readthedocs.io/en/latest/userguide.html
        .. [2] https://esi-acme.readthedocs.io/en/latest/advanced_usage.html
        """

        # First and foremost, set up logging system - logfile is processed later
        prepare_log(logname="ACME", verbose=verbose)
        log.announce("This is ACME v. %s", __version__)                 # type: ignore

        # Backwards compatibility: legacy keywords are converted to new nomenclature
        if any(kw in kwargs for kw in __deprecated__):
            log.warning(__deprecation_wrng__)
            n_workers = kwargs.pop("n_jobs", n_workers)                 # type: ignore
            mem_per_worker = kwargs.pop("mem_per_job", mem_per_worker)  # type: ignore
            log.debug("Set `n_workers = n_jobs` and \
                       mem_per_worker = mem_per_job`")

        # Either guess `n_inputs` or use provided value to duplicate input args
        # and set class attributes `n_inputs`, `argv` and `kwargv`
        self.prepare_input(func, n_inputs, *args, **kwargs)

        # Create an instance of `ACMEdaemon` that does the actual parallel computing work
        log.debug("Instantiating `ACMEdaemon`")
        self.daemon = ACMEdaemon(self,
                                 n_workers=n_workers,
                                 write_worker_results=write_worker_results,
                                 output_dir=output_dir,
                                 result_shape=result_shape,
                                 result_dtype=result_dtype,
                                 single_file=single_file,
                                 write_pickle=write_pickle,
                                 dryrun=dryrun,
                                 partition=partition,
                                 mem_per_worker=mem_per_worker,
                                 setup_timeout=setup_timeout,
                                 setup_interactive=setup_interactive,
                                 stop_client=stop_client,
                                 verbose=verbose,
                                 logfile=logfile)

    def prepare_input(
            self,
            func: Callable,
            n_inputs: Union[int, str],
            *args: Any,
            **kwargs: Optional[Any]) -> None:
        """
        User input parser

        Ensure `func` can actually process provided arguments. If `n_inputs` was
        not set, attempt to infer the number of required concurrent function
        calls from `args` and `kwargs`. In addition, ensure the size of each
        argument is "reasonable" for propagation across multiple workers.
        """

        # Ensure `func` really is a function and `n_inputs` makes sense
        if not callable(func):
            msg = "%s first input has to be a callable function, not %s"
            raise TypeError(msg%(self.objName, str(type(func))))
        if isinstance(n_inputs, str):
            if n_inputs != "auto":
                msg = "%s `n_inputs` has to be 'auto' or an integer >= 2, not %s"
                raise ValueError(msg%(self.objName, n_inputs))
            guessInputs = True
            log.debug("Using `n_inputs = 'auto'`")
        else:
            try:
                _scalar_parser(n_inputs, varname="n_inputs", ntype="int_like", lims=[1, np.inf])
            except Exception as exc:
                log.error("Error parsing `n_inputs`")
                raise exc
            guessInputs = False
            log.debug("Using provided `n_inputs = %d`", n_inputs)

        # Get `func`'s signature to extract its positional/keyword arguments
        funcSignature = inspect.signature(func)
        funcPosArgs = [name for name, value in funcSignature.parameters.items()\
            if value.default is value.empty and value.name != "kwargs"]
        funcKwargs = [name for name, value in funcSignature.parameters.items()\
            if value.default is not value.empty]

        # Account for positional args that were specified by name (keyword-like)
        args = list(args)
        posArgNames = []
        for name, value in kwargs.items():
            if name in funcPosArgs:
                args.insert(funcPosArgs.index(name), value)
                posArgNames.append(name)
        for name in posArgNames:
            kwargs.pop(name)
            log.debug("Moved %s from `kwargs` to positional args", name)

        # If "taskID" is a keyword arg, include/overwrite it in `kwargs` - the rest
        # is done by `ACMEdaemon`
        if "taskID" in funcKwargs:
            kwargs["taskID"] = None
            log.debug("Found `taskID` in kwargs - overriding it")

        # Compare provided `args`/`kwargs` to actually defined quantities in `func`
        if len(args) != len(funcPosArgs):
            msg = "%s %s expects %d positional arguments (%s), found %d"
            validArgs = "'" + "'".join(arg + "', " for arg in funcPosArgs)[:-2]
            raise ValueError(msg%(self.objName,
                                  func.__name__,
                                  len(funcPosArgs),
                                  validArgs,
                                  len(args)))
        if len(kwargs) > len(funcKwargs):
            msg = "%s %s accepts at maximum %d keyword arguments (%s), found %d"
            validArgs = "'" + "'".join(arg + "', " for arg in funcKwargs)[:-2]
            raise ValueError(msg%(self.objName,
                                  func.__name__,
                                  len(funcKwargs),
                                  validArgs,
                                  len(kwargs)))

        # Prepare argument parsing: collect the the length of anything 1D-array-like
        # in `argLens` and check the size of all provided positional and keyword args
        argLens = []
        wrnMsg = "argument size %4.2f MB exceeds recommended limit of %d MB. " +\
            "Distributing large variables across workers may result in poor performance. "

        # Cycle through positional args
        for k, arg in enumerate(args):
            if isinstance(arg, range):
                arg = list(arg)
                args[k] = arg
            acs.callCount = 0
            argsize = sizeOf(arg, "positional arguments")
            log.debug("Computed size of pos arg #%d as %4.2f bytes", k, argsize)
            if argsize > self._maxArgSize:
                log.warning(wrnMsg, argsize, self._maxArgSize)
            if isinstance(arg, (list, tuple)):
                argLens.append(len(arg))
            elif isinstance(arg, np.ndarray):
                if len(arg.squeeze().shape) == 1:
                    argLens.append(arg.squeeze().size)

        # More complicated for keyword args: if provided value is 1D-array-like and
        # default value is not, interpret is as worker-arg list, and track its length
        for name, value in kwargs.items():
            defaultValue = funcSignature.parameters[name].default
            if isinstance(value, range):
                value = list(value)
                kwargs[name] = value
            acs.callCount = 0
            valsize = sizeOf(value, "keyword arguments")
            log.debug("Computed size of %s as %4.2f bytes", name, valsize)
            if valsize > self._maxArgSize:
                log.warning(wrnMsg, valsize, self._maxArgSize)
            if isinstance(value, (list, tuple)):
                if isinstance(defaultValue, (list, tuple)):
                    if len(defaultValue) != len(value):
                        argLens.append(len(value))
                else:
                    argLens.append(len(value))
            elif isinstance(value, np.ndarray) and not isinstance(defaultValue, np.ndarray):
                if len(value.squeeze().shape) == 1:
                    argLens.append(value.squeeze().size)

        # If `n_input` is `"auto"`, make an educated guess as to how many parallel
        # executions of `func` are intended; if input args contained multiple
        # 1D-array-likes, pump the brakes. If `n_input` was explicitly provided,
        # either all input arguments must have unit length (or are nd-arrays)
        # or at at least one input argument actually contains `n_input` elements
        if guessInputs:
            if len(set(argLens)) > 1 or len(argLens) == 0:
                msg = "%s automatic input distribution failed: found %d objects " +\
                    "containing %d to %d elements. Please specify `n_inputs` manually. "
                raise ValueError(msg%(self.objName,
                                      len(argLens),
                                      min(argLens, default=0),
                                      max(argLens, default=0)))
            n_inputs = argLens[0]
        else:
            if n_inputs not in set(argLens) and not all(arglen == 1 for arglen in argLens):
                msg = "%s No object has required length of %d matching `n_inputs`. "
                raise ValueError(msg%(self.objName, n_inputs))
        self.n_inputs = int(n_inputs)
        log.debug("Inferred `n_inputs = %d`", n_inputs)

        # Anything that does not contain `n_input` elements is converted to a one-element list
        wrnMsg = "Found a single callable object in positional arguments. " +\
            "It will be executed just once and shared by all workers"
        self.argv = list(args)
        for ak, arg in enumerate(args):
            if isinstance(arg, (list, tuple)):
                if len(arg) == self.n_inputs:
                    continue
            elif isinstance(arg, np.ndarray):
                if len(arg.squeeze().shape) == 1 and arg.squeeze().size == self.n_inputs:
                    continue
            elif callable(arg):
                log.warning(wrnMsg)
            self.argv[ak] = [arg]

        # Same for keyword arguments with the caveat that default values have to
        # be taken into account (cf. above)
        wrnMsg = "Found a single callable object in keyword arguments: %s. " +\
            "It will be executed just once and shared by all workers"
        self.kwargv = dict(kwargs)
        for name, value in kwargs.items():
            if isinstance(value, (list, tuple)):
                if len(value) == self.n_inputs:
                    continue
            elif isinstance(value, np.ndarray) and \
                not isinstance(funcSignature.parameters[name].default, np.ndarray):
                if len(value.squeeze().shape) == 1 and value.squeeze().size == self.n_inputs:
                    continue
            elif callable(value):
                log.warning(wrnMsg, name)
            self.kwargv[name] = [value]

        # Finally, attach user-provided function to class instance
        self.func = func

        # Get out
        return

    def compute(self) -> None:
        """
        Shortcut to launch parallel computation via `ACMEdaemon`
        """
        log.debug("Invoking `compute` method")
        if hasattr(self, "daemon"):
            self.daemon.compute()
        return

    def cleanup(self) -> None:
        """
        Shortcut to corresponding cleanup-routine provided by `ACMEdaemon`
        """
        log.debug("Invoking `cleanup` method")
        if hasattr(self, "daemon"):
            self.daemon.cleanup()
        return

    def __enter__(self) -> ACMEdaemon:
        """
        If `ParallelMap` is used as context manager, launch `ACMEdaemon`
        """
        log.debug("Entering `ACMEdaemon` context")
        return self.daemon

    def __exit__(
            self,
            exception_type: Any,
            exception_value: Any,
            exception_traceback: Any) -> None:
        """
        If `ParallelMap` is used as context manager, close any ad-hoc computing
        clients created by `ACMEdaemon`
        """
        log.debug("Exiting `ACMEdaemon` context")
        self.daemon.cleanup()
        return



#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtins
import datetime
from setuptools import setup
from setuptools.config.setupcfg import read_configuration
import toml

# External packages
import yaml

# Manually set release version
releaseVersion = "2023.12"

# Read dependencies from setup.cfg + pyproject.toml and create conda environment file
envFile = "acme.yml"
setupOpts = read_configuration("setup.cfg")["options"]
tomlPkgs = toml.load("pyproject.toml")["build-system"]["requires"]
allPkgs = setupOpts["install_requires"] + setupOpts["extras_require"]["dev"] + tomlPkgs
pipPkgs = ["sphinx_automodapi"]
for k in range(len(pipPkgs)):
    pkg = pipPkgs[k]
    pipPkgs[k] = allPkgs.pop([allPkgs.index(dep) for dep in allPkgs if pkg in dep][0])
condaMap = {"msgpack" : "msgpack-python"}
for pipName, condaName in condaMap.items():
    depIdx = [allPkgs.index(dep) for dep in allPkgs if pipName in dep][0]
    depName = allPkgs[depIdx]
    allPkgs[depIdx] = depName.replace(pipName, condaName)
pyVer = setupOpts
allPkgs += ["python " + str(setupOpts["python_requires"]), "pip", {"pip" : pipPkgs}]
ymlDict = {"name" : "acme",
           "channels" : ["defaults", "conda-forge"],
           "dependencies" : allPkgs}
header = "#\n" +\
    "# Copyright © {} Ernst Strüngmann Institute (ESI) for Neuroscience " +\
    "# in Cooperation with Max Planck Society\n#\n" +\
    "# SPDX-License-Identifier: CC0-1.0\n#\n" +\
    "# This file was auto-generated by setup.py on {}. \n" +\
    "# Do not edit, all of your changes will be overwritten. \n"
header = header.format(datetime.datetime.now().strftime("%Y"),
                       datetime.datetime.now().strftime("%d/%m/%Y at %H:%M:%S"))
with open(envFile, "w", encoding="utf8") as ymlFile:
    ymlFile.write(header)
    yaml.dump(ymlDict, ymlFile, default_flow_style=False)

# Update citation file
citationFile = "CITATION.cff"
with open(citationFile, "r", encoding="utf8") as ymlFile:
    ymlObj = yaml.safe_load(ymlFile)
ymlObj["version"] = releaseVersion
ymlObj["date-released"] = datetime.datetime.now().strftime("%Y-%m-%d")
with open(citationFile, "w", encoding="utf8") as ymlFile:
    yaml.dump(ymlObj, ymlFile)

# Run setup (note: identical arguments supplied in setup.cfg will take precedence)
setup(version=releaseVersion)



#
# Computational scaffolding for user-interface
#
# Copyright © 2023 Ernst Strüngmann Institute (ESI) for Neuroscience
# in Cooperation with Max Planck Society
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Builtin/3rd party package imports
import time
import socket
import getpass
import datetime
import inspect
import numbers
import collections
import os
import sys
import glob
import shutil
import functools
import pickle
import logging
import multiprocessing
import psutil
import tqdm
import h5py
import dask
import dask.distributed as dd
import dask_jobqueue as dj
import numpy as np
from typing import TYPE_CHECKING, Optional, Any, Union, List
from numpy.typing import ArrayLike

# Local imports
from . import __path__
from .dask_helpers import (esi_cluster_setup, local_cluster_setup,
                           slurm_cluster_setup, cluster_cleanup, count_online_workers)
from .shared import user_yesno, is_esi_node, is_slurm_node
from .logger import prepare_log
isSpyModule = False
if "syncopy" in sys.modules:            # pragma: no cover
    isSpyModule = True
if TYPE_CHECKING:                       # pragma: no cover
    from frontend import ParallelMap

__all__: List["str"] = ["ACMEdaemon"]

# Fetch logger
log = logging.getLogger("ACME")


# Main manager for parallel execution of user-defined functions
class ACMEdaemon(object):

    # Restrict valid class attributes
    __slots__ = "func", "argv", "kwargv", "n_calls", "n_workers", "acme_func", \
        "task_ids", "out_dir", "collect_results", "results_container", "result_shape", \
        "result_dtype", "stacking_dim", "client", "stop_client", "has_slurm"

    # Prepend every stdout/stderr message with the name of this class
    objName = "<ACMEdaemon>"

    # format string for tqdm progress bars
    tqdmFormat = "{desc}: {percentage:3.0f}% |{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}]"

    # time (in seconds) b/w querying state of futures ('pending' -> 'finished')
    sleepTime = 0.1

    def __init__(
        self,
        pmap: "ParallelMap",
        n_workers: Union[int, str] = "auto",
        write_worker_results: bool = True,
        output_dir: Optional[str] = None,
        result_shape: Optional[tuple[Optional[int], ...]] = None,
        result_dtype: str = "float",
        single_file: bool = False,
        write_pickle: bool = False,
        dryrun: bool = False,
        partition: str = "auto",
        mem_per_worker: str = "auto",
        setup_timeout: int = 60,
        setup_interactive: bool = True,
        stop_client: Union[bool, str] = "auto",
        verbose: Optional[bool] = None,
        logfile: Optional[Union[bool, str]] = None) -> None:
        """
        Manager class for performing concurrent user function calls

        Parameters
        ----------
        pmap : :class:`~acme.ParallelMap` context manager
            By default, `:class:~`acme.ACMEDaemon` assumes  that that
            the provided :class:`~acme.ParallelMap` instance has already
            been properly set up to process `func` (all input arguments parsed and
            properly formatted). All other input arguments of `:class:~`acme.ACMEDaemon`
            are extracted from the provided :class:`~acme.ParallelMap` instance.
        n_workers : int or "auto"
            Number of SLURM workers (=jobs) to spawn. See :class:`~acme.ParallelMap`
            for details.
        write_worker_results : bool
            If `True`, the return value(s) of `func` is/are saved on disk. See
            :class:`~acme.ParallelMap` for details.
        output_dir : str or None
            If provided, auto-generated results are stored in the given path. See
            :class:`~acme.ParallelMap` for details.
        result_shape : tuple or None
            If provided, results are slotted into a dataset/array with layout `result_shape`. See
            :class:`~acme.ParallelMap` for details.
        result_dtype : str
            Determines numerical datatype of dataset laid out by `result_shape`.
            See :class:`~acme.ParallelMap` for details.
        single_file : bool
            If `True`, parallel workers write to the same results container. See
            :class:`~acme.ParallelMap` for details.
        write_pickle : bool
            If `True`, the return value(s) of `func` is/are pickled to disk. See
            :class:`~acme.ParallelMap` for details.
        dryrun : bool
            If `True`, a dry-run of calling `func` is performed using a single
            `args`, `kwargs` tuple. See :class:`~acme.ParallelMap` for details.
        partition : str
            Name of SLURM partition to use. See :class:`~acme.ParallelMap` for details.
        mem_per_worker : str
            Memory booking for each SLURM worker. See :class:`~acme.ParallelMap` for details.
        setup_timeout : int
            Timeout period (in seconds) for SLURM workers to come online. See
            :class:`~acme.ParallelMap` for details.
        setup_interactive : bool
            If `True`, user input is queried in case not enough SLURM workers could
            be started within `setup_timeout` seconds. See :class:`~acme.ParallelMap`
            for details.
        stop_client : bool or "auto"
            If `"auto"`, automatically started distributed computing clients
            are shut down at the end of computation, while user-provided clients
            are left untouched. See :class:`~acme.ParallelMap` for details.
        verbose : None or bool
            If `None` (default), general run-time information as well as warnings
            and errors are shown. See :class:`~acme.ParallelMap` for details.
        logfile : None or bool or str
            If `None` (default) or `True`, and `write_worker_results` is
            `True`, all run-time information as well as errors and
            warnings are tracked in a log-file. See :class:`~acme.ParallelMap`
            for details.

        Returns
        -------
        results : list
            If `write_worker_results` is `True`, `results` is a list of HDF5 file-names
            containing computed results. If `write_worker_results` is `False`,
            results is a list comprising the actual return values of `func`.
            If `:class:~`acme.ACMEDaemon` was instantiated by :class:`~acme.ParallelMap`,
            results are propagated back to :class:`~acme.ParallelMap`.

        See also
        --------
        ParallelMap : Context manager and main user interface
        """

        # The only error checking happening in `__init__`
        try:
            pClassName = pmap.__class__.__name__
        except:
            pClassName = "notParallelMap"
        if pClassName != "ParallelMap":
            msg = "%s `pmap` has to be a `ParallelMap` instance, not %s"
            raise TypeError(msg%(self.objName, str(type(pmap))))

        # Allocate first batch of slots
        self.func = pmap.func
        self.argv = pmap.argv
        self.kwargv = pmap.kwargv
        self.n_calls = pmap.n_inputs
        self.n_workers = None
        self.acme_func = None
        self.out_dir = None
        self.collect_results = None
        self.results_container = None
        self.result_shape = None
        self.result_dtype = None
        self.stacking_dim = None
        self.client = None
        self.stop_client = None

        # Define list of taskIDs for distribution across workers
        self.task_ids = list(range(self.n_calls))
        log.debug("Allocated `taskID` list: %s", str(self.task_ids))

        # Finally, determine if the code is executed on a SLURM-enabled node
        self.has_slurm = is_slurm_node()
        log.debug("Set `has_slurm = %s`", str(self.has_slurm))

        # Set up output handler
        self.pre_process(verbose,
                         logfile,
                         write_worker_results,
                         output_dir,
                         result_shape,
                         result_dtype,
                         single_file,
                         write_pickle)

        # If requested, perform single-worker dry-run (and quit if desired)
        if dryrun:
            goOn = self.perform_dryrun(setup_interactive)
            if not goOn:
                log.debug("Quitting after dryrun")
                return
            log.debug("Continuing after dryrun")

        # Either use existing dask client or start a fresh instance
        self.prepare_client(n_workers=n_workers,
                            partition=partition,
                            mem_per_worker=mem_per_worker,
                            setup_timeout=setup_timeout,
                            setup_interactive=setup_interactive,
                            stop_client=stop_client)

    def pre_process(
            self,
            verbose: Union[bool, None],
            logfile: Union[bool, str, None],
            write_worker_results: bool,
            output_dir: Union[str, None],
            result_shape: Union[tuple[Optional[int], ...], None],
            result_dtype: str,
            single_file: bool,
            write_pickle: bool) -> None:
        """
        If `write_*` is `True` set up directories for saving output HDF5 containers
        (or pickle files). Warn if results are to be collected in memory
        """

        # Basal sanity check for Boolean flags
        if not isinstance(write_worker_results, bool):
            msg = "%s `write_worker_results` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.objName, str(write_worker_results)))
        log.debug("Found `write_worker_results = %s`", str(write_worker_results))
        if not isinstance(single_file, bool):
            msg = "%s `single_file` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.objName, str(single_file)))
        log.debug("Found `single_file = %s`", str(single_file))
        if not isinstance(write_pickle, bool):
            msg = "%s `write_pickle` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.objName, str(write_pickle)))
        log.debug("Found `write_pickle = %s`", str(write_pickle))

        # Check compatibility of provided optional args
        if not write_worker_results and write_pickle:
            log.warning("Pickling of results only possible if `write_worker_results` is `True`. ")
        if not write_worker_results and output_dir:
            log.warning("Output directory specification has no effect if `write_worker_results` is `False`.")
        if write_pickle and result_shape:
            log.warning("Pickling of results does not support output array shape specification. ")
        if not write_worker_results and single_file:
            log.warning("Generating a single output file only possible if `write_worker_results` is `True`. ")
        if write_pickle and single_file:
            msg = "%s Pickling of results does not support single output file creation. "
            raise ValueError(msg%self.objName)

        # Check validity of output shape/dtype specifications
        if result_shape is not None:
            if not isinstance(result_shape, (list, tuple)):
                msg = "%s `result_shape` has to be either `None` or tuple, not %s"
                raise TypeError(msg%(self.objName, str(type(result_shape))))

            if not isinstance(result_dtype, str):
                msg = "%s `result_dtype` has to be a string, not %s"
                raise TypeError(msg%(self.objName, str(type(result_shape))))

            if sum(spec is None for spec in result_shape) != 1:
                msg = "%s `result_shape` must contain exactly one `None` entry"
                raise ValueError(msg%self.objName)

            rShape = list(result_shape)
            self.stacking_dim = result_shape.index(None)                # type: ignore
            rShape[self.stacking_dim] = self.n_calls                    # type: ignore
            if write_worker_results:
                self.kwargv["stackingDim"] = [self.stacking_dim]

            if not all(isinstance(spec, numbers.Number) for spec in rShape):
                msg = "%s `result_shape` must only contain numerical values"
                raise ValueError(msg%self.objName)
            if any(spec < 0 or int(spec) != spec for spec in rShape):   # type: ignore
                msg = "%s `result_shape` must only contain non-negative integers"
                raise ValueError(msg%self.objName)

            self.result_shape = tuple(rShape)                           # type: ignore
            msg = "Found `result_shape = %s`. Set stacking dimension to %d"
            log.debug(msg, str(result_shape), self.stacking_dim)

            try:
                self.result_dtype = np.dtype(result_dtype)              # type: ignore
            except Exception as exc:
                msg = "%s `result_dtype` has to be a valid NumPy datatype specification. "
                msg += "Original error message below:\n%s"
                raise TypeError(msg%(self.objName, str(exc)))
            log.debug("Set `result_dtype = %s`", self.result_dtype)
        else:
            log.debug("Found `result_shape = %s`", str(result_shape))
            log.debug("Found `result_dtype = %s`", str(result_dtype))

        # If automatic saving of results is requested, make necessary preparations
        if write_worker_results:
            self.setup_output(output_dir, result_shape, single_file, write_pickle)
        else:

            # If `taskID` is not an explicit kw-arg of `func` and `func` does not
            # accept "anonymous" `**kwargs`, don't save anything but return stuff
            log.debug("Automatic output processing disabled.")
            if self.kwargv.get("taskID") is None:
                if not isSpyModule:
                    msg = "`write_worker_results` is `False` and `taskID` is not a keyword argument of %s. " +\
                        "Results will be collected in memory by caller - this might be slow and can lead " +\
                        "to excessive memory consumption. "
                    log.warning(msg, self.func.__name__)
                self.collect_results = True                             # type: ignore
            else:
                self.kwargv["taskID"] = self.task_ids
                self.collect_results = False                            # type: ignore
                msg = "Not collecting results in memory, leaving output " +\
                    "processing to user-provided function"
                log.debug(msg)

            # The "raw" user-provided function is used in the computation
            self.acme_func = self.func
            log.debug("Not wrapping user-provided function but invoking it directly")

        # Unless specifically disabled by the user, enable progress-tracking
        # in a log-file if results are auto-generated
        if logfile is None and write_worker_results is True:
            logfile = True

        # Either parse provided `logfile` or set up an auto-generated file;
        # After this test, `logfile` is either a filename or `None`
        msg = "%s `logfile` has to be `None`, `True`, `False` or a valid file-name, not %s"
        if logfile is None or isinstance(logfile, bool):
            if logfile is True:
                if write_worker_results:
                    logfile = self.out_dir
                else:
                    logfile = os.path.dirname(os.path.abspath(inspect.getfile(self.func)))
                logfile = os.path.join(logfile, "ACME_{func:s}_{date:s}.log")   # type: ignore
                logfile = logfile.format(func=self.func.__name__,
                                         date=datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))
            else:
                logfile = None
        elif isinstance(logfile, str):
            if os.path.isdir(logfile):
                raise IOError(msg%(self.objName, "a directory"))
            logfile = os.path.abspath(os.path.expanduser(logfile))
        else:
            raise TypeError(msg%(self.objName, str(type(logfile))))

        # If progress tracking in a log-file was requested, set it up now
        prepare_log(logname="ACME", logfile=logfile, verbose=verbose)
        log.debug("Set up logfile=%s", str(logfile))

        return

    def setup_output(
            self,
            output_dir: Union[str, None],
            result_shape: Union[tuple[Optional[int], ...], None],
            single_file: bool,
            write_pickle: bool) -> None:
        """
        Local helper for creating output directories and preparing containers
        """

        # Check validity of output dir specification
        if not isinstance(output_dir, (type(None), str)):
            msg = "%s `output_dir` has to be either `None` or str, not %s"
            raise TypeError(msg%(self.objName, str(type(output_dir))))
        log.debug("Found `output_dir = %s`", str(output_dir))

        # If provided, standardize output dir spec, otherwise use default locations
        if output_dir is not None:
            outDir = os.path.abspath(os.path.expanduser(output_dir))

        else:
            # On the ESI cluster, save results on HPC mount, otherwise use location of `func`
            if self.has_slurm:
                outDir = "/cs/home/{usr:s}/".format(usr=getpass.getuser())
            else:                                                       # pragma: no cover
                outDir = os.path.dirname(os.path.abspath(inspect.getfile(self.func)))
            outDir = os.path.join(outDir, "ACME_{date:s}")
            outDir = outDir.format(date=datetime.datetime.now().strftime('%Y%m%d-%H%M%S-%f'))
        log.debug("Using output directory %s", outDir)

        # Unless specifically denied by the user, each worker stores results
        # separately with a common container file pointing to the individual
        # by-worker files residing in a "payload" directory
        self.out_dir = str(outDir)                                      # type: ignore
        if not single_file and not write_pickle:
            log.debug("Preparing payload directory for HDF5 containers")
            payloadName = "{}_payload".format(self.func.__name__)
            outputDir = os.path.join(self.out_dir, payloadName)         # type: ignore
        else:
            msg = "Either single-file output or pickling was requested. " +\
                "Not creating payload directory"
            log.debug(msg)
            outputDir = self.out_dir
        try:
            os.makedirs(outputDir)
            log.debug("Created %s", outputDir)
        except Exception as exc:
            msg = "%s automatic creation of output folder %s failed: %s"
            raise OSError(msg%(self.objName, outputDir, str(exc)))

        # Re-define or allocate key "taskID" to track concurrent processing results
        self.kwargv["taskID"] = self.task_ids
        self.collect_results = False                                    # type: ignore

        # Set up correct file-extension for output files; in case of HDF5
        # containers, prepare "main" file for collecting/symlinking worker results
        if write_pickle:
            fExt = "pickle"
            log.debug("Pickling was requested")
        else:
            fExt = "h5"
            self.results_container = os.path.join(self.out_dir, "{}.h5".format(self.func.__name__))     # type: ignore
            log.debug("Using HDF5 storage %s", self.results_container)

        # By default, `results_container` is a collection of links that point to
        # worker-generated HDF5 containers; if `single_file` is `True`, then
        # `results_container` is a "real" container with actual dataset(s)
        if single_file:
            self.kwargv["singleFile"] = [True]
            self.kwargv["outFile"] = [self.results_container]
            log.debug("Saving results in single HDF5 container")

            # If no output shape provided, prepare groups for storing datasets;
            # otherwise allocate a single dataset w/specified dimension
            if result_shape is None:
                msg = "Created group comp_%d in single shared results container"
                with h5py.File(self.results_container, "w") as h5f:
                    for i in self.task_ids:
                        h5f.create_group("comp_{}".format(i))
                        log.debug(msg, i)
            else:
                msg = "Created unique dataset 'result_0' with shape %s " +\
                    "in single shared results container"
                with h5py.File(self.results_container, "w") as h5f:
                    h5f.create_dataset("result_0",
                                       shape=self.result_shape,
                                       dtype=self.result_dtype)
                    log.debug(msg, str(self.result_shape))

        else:
            self.kwargv["outFile"] = [os.path.join(outputDir,
                                                    "{}_{}.{}".format(self.func.__name__,
                                                                        taskID,
                                                                        fExt))
                                                    for taskID in self.task_ids]
            if not write_pickle:

                # If no output shape provided, generate links to external datasets;
                # otherwise allocate a virtual dataset w/specified dimension
                if result_shape is None:

                    msg = "Created external link comp_%d pointing to " +\
                        "%s in results container"
                    with h5py.File(self.results_container, "w") as h5f:
                        for i, fname in enumerate(self.kwargv["outFile"]):
                            relPath = os.path.join(payloadName, os.path.basename(fname))
                            h5f["comp_{}".format(i)] = h5py.ExternalLink(relPath, "/")
                            log.debug(msg, i, relPath)
                else:

                    # Assemble virtual dataset
                    ShapeSource = list(self.result_shape)               # type: ignore
                    ShapeSource.pop(self.stacking_dim)
                    ShapeSource = tuple(ShapeSource)
                    layout = h5py.VirtualLayout(shape=self.result_shape, dtype=self.result_dtype)   # type: ignore
                    idx = [slice(None)] * len(self.result_shape)        # type: ignore
                    msg = "Created virtual dataset result_0' with shape " +\
                        "%s in results container"
                    for i, fname in enumerate(self.kwargv["outFile"]):
                        idx[self.stacking_dim] = i                      # type: ignore
                        relPath = os.path.join(payloadName, os.path.basename(fname))
                        vsource = h5py.VirtualSource(fname, "result_0", shape=ShapeSource)
                        layout[tuple(idx)] = vsource
                    with h5py.File(self.results_container, "w", libver="latest") as h5f:
                        h5f.create_virtual_dataset("result_0", layout)
                        log.debug(msg, self.result_shape)

        # Include logger name in keywords so that workers can use it
        self.kwargv["logName"] = [log.name]

        # Wrap the user-provided func and distribute it across workers
        self.kwargv["userFunc"] = [self.func]
        self.acme_func = self.func_wrapper                              # type: ignore
        log.debug("Wrapping user-provided function inside func_wrapper")

        # Finally, attach verbosity flag to enable logging inside wrapper
        self.kwargv["logLevel"] = [log.level]

        return

    def perform_dryrun(
            self,
            setup_interactive: bool) -> bool:
        """
        Execute user function with one prepared randomly picked args, kwargs combo
        """

        # Let helper randomly pick a single scheduled job and prepare corresponding args + kwargs
        [dryRunIdx], [dryRunArgs], [dryRunKwargs] = self._dryrun_setup(n_runs=1)    # type: ignore

        # Create log entry
        msg = "Performing a single dry-run of %s simulating randomly " +\
            "picked worker #%d with automatically distributed arguments"
        log.info(msg, self.func.__name__, dryRunIdx)

        # Use resident memory size (in MB) to estimate job's memory footprint and measure elapsed time
        mem0 = psutil.Process().memory_info().rss / 1024 ** 2
        log.debug("Initial memory consumption estimate: %3.f MB", mem0)
        log.debug("Starting dryrun")
        tic = time.perf_counter()
        self.acme_func(*dryRunArgs, **dryRunKwargs)                     # type: ignore
        toc = time.perf_counter()
        log.debug("Finished dryrun")
        mem1 = psutil.Process().memory_info().rss / 1024 ** 2
        log.debug("Memory consumption estimate after dryrun: %3.f MB", mem1)

        # Remove any generated output files
        if self.out_dir is not None:
            log.debug("Removing %s generated during dryrun", self.kwargv["outFile"][dryRunIdx])
            os.unlink(self.kwargv["outFile"][dryRunIdx])

        # Compute elapsed time and memory usage
        elapsedTime = toc - tic
        memUsage = mem1 - mem0

        # Prepare info message
        memUnit = "MB"
        if memUsage > 1000:
            memUsage /= 1024
            memUnit = "GB"
        msg = "Dry-run completed. Elapsed time is %f seconds, " +\
            "estimated memory consumption was %3.2f %s."
        log.info(msg, elapsedTime, memUsage, memUnit)

        # If the worker setup is supposed to be interactive, ask for confirmation
        # here as well; if execution is terminated, remove auto-generated output directory
        goOn = True
        if setup_interactive:
            msg = "Do you want to continue executing {fname:s} with the provided arguments?"
            if not user_yesno(msg.format(fname=self.func.__name__), default="yes"):
                if self.out_dir is not None:
                    shutil.rmtree(self.out_dir, ignore_errors=True)
                goOn = False
        return goOn

    def _dryrun_setup(
            self,
            n_runs: Optional[int] = None) -> tuple[ArrayLike, List, List]:
        """
        Pick scheduled job(s) at random and extract corresponding (already prepared!) args + kwargs
        """

        # If not provided, attempt to infer a "sane" default for the number of jobs to pick
        if n_runs is None:
            n_runs = min(self.n_calls, max(5, min(1, int(0.05 * self.n_calls))))
        log.debug("Picking %d jobs at random", n_runs)

        # Randomly pick `n_runs` jobs and extract positional and keyword args
        dryRunIdx = np.random.choice(self.n_calls, size=n_runs, replace=False)
        dryRunArgs = []
        dryRunKwargs = []
        for idx in dryRunIdx:
            dryRunArgs.append([arg[idx] if len(arg) > 1 else arg[0] for arg in self.argv])
            dryRunKwargs.append([{key:value[idx] if len(value) > 1 else value[0] \
                for key, value in self.kwargv.items()}][0])
        return dryRunIdx, dryRunArgs, dryRunKwargs

    def prepare_client(
            self,
            n_workers: Union[int, str] = "auto",
            partition: str = "auto",
            mem_per_worker: str = "auto",
            setup_timeout: int = 60,
            setup_interactive: bool = True,
            stop_client: Union[bool, str] = "auto") -> None:
        """
        Setup or fetch dask distributed processing client. Depending on available
        hardware, either start a local multi-processing client or launch a
        worker cluster via SLURM.

        Also ensure that ad-hoc clients created here are stopped and worker jobs
        are properly released at the end of computation. However, ensure any client
        not created by `prepare_client` is **not** automatically cleaned up.
        """

        # Modify automatic setting of `stop_client` if requested
        msg = "%s `stop_client` has to be 'auto' or Boolean, not %s"
        if isinstance(stop_client, str):
            if stop_client != "auto":
                raise ValueError(msg%(self.objName, stop_client))
        elif isinstance(stop_client, bool):
            self.stop_client = stop_client                              # type: ignore
        else:
            raise TypeError(msg%(self.objName, str(type(stop_client))))
        log.debug("Using `stop_client = %s`", str(stop_client))

        # Check if a dask client is already running
        try:
            self.client = dd.get_client()                               # type: ignore
            log.debug("Detected running client %s", str(self.client))
            if stop_client == "auto":
                self.stop_client = False                                # type: ignore
                msg = "%s Changing `stop_client` from `'auto'` to `False` " +\
                    "to not terminate external client"
                log.debug(msg)
            self.n_workers = count_online_workers(self.client.cluster)  # type: ignore
            log.debug("Found %d alive workers in the client", self.n_workers)
            msg = "Attaching to parallel computing client %s"
            log.info(msg%(str(self.client)))
            return
        except ValueError:
            msg = "No running client detected, preparing to start a new one"
            log.debug(msg)
            if stop_client == "auto":
                self.stop_client = True                                 # type: ignore
                msg = "Changing `stop_client` from `'auto'` to `True` " +\
                    "to clean up client started by `ParallelMap`"
                log.debug(msg)

            # If `n_workers` is `"auto`, set `n_workers = n_calls` (default)
            msg = "%s `n_workers` has to be 'auto' or an integer >= 1, not %s"
            if isinstance(n_workers, str):
                if n_workers != "auto":
                    raise ValueError(msg%(self.objName, n_workers))
                if self.has_slurm:
                    n_workers = self.n_calls
                else:
                    n_workers = None
                log.debug("Changing `n_workers` from `'auto'` to %s", str(n_workers))
            log.debug("Using provided `n_workers = %d` to start client", n_workers)

        # If things are running locally, simply fire up a dask-distributed client,
        # otherwise go through the motions of preparing a full worker cluster
        if not self.has_slurm:                                          # pragma: no cover

            log.debug("SLURM not found, Calling `local_cluster_setup`")
            self.client = local_cluster_setup(n_workers=n_workers, interactive=False)        # type: ignore

        else:

            # If `partition` is "auto", use `estimate_memuse` to heuristically determine
            # average memory consumption of jobs
            log.debug("SLURM available parsing settings")
            if not isinstance(partition, str):
                msg = "%s `partition` has to be 'auto' or a valid SLURM partition name, not %s"
                raise TypeError(msg%(self.objName, str(type(partition))))
            if partition == "auto":
                if is_esi_node():
                    msg = "Automatic SLURM partition selection is experimental"
                    log.warning(msg)
                    mem_per_worker = self.estimate_memuse()
                else:                                                   # pragma: no cover
                    err = "Automatic SLURM partition selection currently only available " +\
                        "on the ESI HPC cluster. "
                    log.error(err)

            # All set, remaining input processing is done by respective `*_cluster_setup` routines
            if is_esi_node():
                msg = "Running on ESI compute node, Calling `esi_cluster_setup`"
                log.debug(msg)
                self.client = esi_cluster_setup(partition=partition, n_workers=n_workers,               # type: ignore
                                                mem_per_worker=mem_per_worker, timeout=setup_timeout,
                                                interactive=setup_interactive, start_client=True)

            # Unknown cluster node, use vanilla config
            else:                                                       # pragma: no cover
                wrng = "Cluster node %s not recognized. Falling back to vanilla " +\
                    "SLURM setup allocating one worker and one core per worker"
                log.warning(wrng%(socket.getfqdn()))
                processes_per_worker = 1
                n_cores = 1
                self.client = slurm_cluster_setup(partition=partition,                                  # type: ignore
                                                  n_cores=n_cores,
                                                  n_workers=n_workers,                                  # type: ignore
                                                  processes_per_worker=processes_per_worker,
                                                  mem_per_worker=mem_per_worker,
                                                  n_workers_startup=1,
                                                  timeout=setup_timeout,
                                                  interactive=setup_interactive,
                                                  interactive_wait=120,
                                                  start_client=True,
                                                  job_extra=[],
                                                  invalid_partitions=[])

            # If startup is aborted by user, get outta here
            if self.client is None:                                     # pragma: no cover
                msg = "%s Could not start distributed computing client. "
                raise ConnectionAbortedError(msg%(self.objName))

        # Set `n_workers` to no. of active workers in the initialized cluster
        self.n_workers = len(self.client.cluster.workers)               # type: ignore
        log.debug("Setting `n_workers = %d` based on active workers in %s",
                  self.n_workers, str(self.client))

        # If single output file saving was chosen, initialize distributed
        # lock for shared writing to container
        if self.kwargv.get("singleFile") is not None:
            msg = "Initializing distributed lock for writing to single shared results container"
            log.debug(msg)
            dd.lock.Lock(name=os.path.basename(self.results_container)) # type: ignore

        return

    def estimate_memuse(self) -> str:
        """
        A brute-force guessing approach to determine memory consumption of provided
        workload
        """

        # Let helper randomly pick some jobs and prepare corresponding args + kwargs
        dryRunIdx, dryRunArgs, dryRunKwargs = self._dryrun_setup()

        # Append new dummy keyword to return before any disk-writes happen
        # in case ACME handles results output
        if self.out_dir is not None:
            msg = "Appending `memEstRun` keyword to func_wrapper to prevent any disk-writes"
            log.debug(msg)
            for k in range(len(dryRunKwargs)):
                dryRunKwargs[k]["memEstRun"] = True

        # Set run-time for each job (in seconds) and allocate arrays for memory
        # consumption estimates (in GB)
        runTime = 30
        memPerSec = np.zeros((runTime,))
        memPerJob = np.zeros((len(dryRunIdx),))                         # type: ignore

        # Adequately warn about this heuristic gymnastics...
        log.info("Estimating memory footprint of %s", self.func.__name__)

        msg = "Running %d random workers evaluating %s for max. %d seconds"
        log.debug(msg%(len(dryRunIdx), self.func.__name__, runTime))    # type: ignore
        wmsg = "Launching worker #{wrknum:d}"
        for i, idx in enumerate(dryRunIdx):                             # type: ignore

            # Set up dedicated process to execute user-provided function w/allocated args/kwargs
            proc = multiprocessing.Process(target=self.acme_func,
                                           args=dryRunArgs[i],
                                           kwargs=dryRunKwargs[i])

            # Run user-func for max. `runTime` seconds (or worker finishes),
            # get memory footprint every second
            proc.start()
            with tqdm.tqdm(desc=wmsg.format(wrknum=idx),
                           total=runTime,
                           bar_format=self.tqdmFormat,
                           position=0) as pbar:
                for k in range(runTime):
                    memPerSec[k] = psutil.Process(proc.pid).memory_info().rss / 1024 ** 3
                    time.sleep(1)
                    pbar.update(1)
                    if not proc.is_alive():
                        pbar.n = runTime
                        break
            proc.kill()

            # Compute peak memory consumption across `runTime` seconds
            memPerJob[i] = memPerSec.max()
            log.debug("Peak memory for worker #%d: %3.2f GB", idx, memPerJob[i])

        # Compute aggregate average memory consumption across all runs
        memUsage = memPerJob.mean()

        # Communicate results
        msg = "Estimated memory consumption across %d runs is %3.2f GB "
        log.info(msg%(len(dryRunIdx), memUsage))                        # type: ignore

        # Return specially formatted string
        mem_per_worker = "estimate_memuse:" + str(max(1, int(np.ceil(memUsage))))
        log.debug("Finished memory estimation, returning `mem_per_worker = %s`", mem_per_worker)
        return mem_per_worker

    def compute(
            self,
            debug: bool = False) -> Union[List, None]:
        """
        Perform the actual parallel execution of `func`

        If `debug` is `True`, use a single-threaded dask scheduler that does
        not actually process anything concurrently but uses the dask framework
        in a sequential setup.
        """

        # If `prepare_client` has not been called yet, don't attempt to compute anything
        if self.client is None:
            log.debug("No parallel computing client allocated, exiting")
            return None

        # Ensure `debug` is a simple Boolean flag
        if not isinstance(debug, bool):
            msg = "%s `debug` has to be `True` or `False`, not %s"
            raise TypeError(msg%(self.objName, str(type(debug))))
        log.debug("Found `debug = %s`", str(debug))

        # Check if the underlying parallel computing cluster hosts actually usable workers
        if count_online_workers(self.client.cluster) == 0:
            msg = "%s no active workers found in distributed computing client %s " +\
                "Consider running \n" +\
                "\timport dask.distributed as dd; dd.get_client().restart()\n" +\
                "If this fails to make workers come online, please use\n" +\
                "\timport acme; acme.cluster_cleanup()\n" +\
                "to shut down any defunct distributed computing clients"
            raise RuntimeError(msg%(self.objName, self.client))
        log.debug("Found %d workers in client %s",
                  count_online_workers(self.client.cluster), str(self.client))

        # Dask does not correctly forward the `sys.path` from the parent process
        # to its workers. Fix this.
        def init_acme(dask_worker, syspath):
            sys.path = list(syspath)
        self.client.register_worker_callbacks(setup=functools.partial(init_acme, syspath=sys.path))
        log.debug("Registered worker callback to forward `sys.path`")

        # Format positional arguments for worker-distribution: broadcast all
        # inputs that are used by all workers and create a list of references
        # to this (single!) future on the cluster for submission
        for ak, arg in enumerate(self.argv):
            if len(arg) == 1:
                ftArg = self.client.scatter(arg, broadcast=True)
                log.debug("Broadcasting single-element pos arg %s to client", str(arg))
                if isinstance(ftArg, collections.abc.Sized):
                    ftArg = ftArg[0]
                self.argv[ak] = [ftArg] * self.n_calls

        # Same as above but for keyword-arguments
        for name, value in self.kwargv.items():
            if len(value) == 1:
                ftVal = self.client.scatter(value, broadcast=True)[0]
                self.kwargv[name] = [ftVal] * self.n_calls
                log.debug("Broadcasting single-element kwarg `%s` to client", name)

        # Re-format keyword arguments to be usable with single-to-many arg submission.
        # Idea: with `self.n_calls = 3` and ``self.kwargv = {'a': [5, 5, 5], 'b': [6, 6, 6]}``
        # then ``kwargList = [{'a': 5, 'b': 6}, {'a': 5, 'b': 6}, {'a': 5, 'b': 6}]``
        kwargList = []
        kwargKeys = self.kwargv.keys()
        kwargVals = list(self.kwargv.values())
        for nc in range(self.n_calls):
            kwDict = {}
            for kc, key in enumerate(kwargKeys):
                kwDict[key] = kwargVals[kc][nc]
            kwargList.append(kwDict)

        # In case a debugging run is performed, use the single-threaded scheduler and return
        if debug:
            log.warning("Running in debug mode")
            with dask.config.set(scheduler='single-threaded'):
                log.debug("Using single-threaded scheduler to evaluate function")
                values = self.client.gather([self.client.submit(self.acme_func, *args, **kwargs) \
                    for args, kwargs in zip(zip(*self.argv), kwargList)])
                return values

        # Depending on the used dask cluster object, point to respective log info
        if isinstance(self.client.cluster, dj.SLURMCluster):
            logFiles = self.client.cluster.job_header.split("--output=")[1].replace("%j", "{}")
            logDir = os.path.split(logFiles)[0]
        else:                                                           # pragma: no cover
            logFiles = []
            logDir = os.path.dirname(self.client.cluster.dashboard_link) + "/info/main/workers.html"
        msg = "Preparing %d parallel calls of `%s` using %d workers"
        log.info(msg%(self.n_calls, self.func.__name__, self.n_workers))
        msg = "Log information available at %s"
        log.debug(msg%(logDir))

        # Submit `self.n_calls` function calls to the cluster
        log.debug("Submitting %d function calls to client %s", self.n_calls, str(self.client))
        futures = [self.client.submit(self.acme_func, *args, **kwargs) \
            for args, kwargs in zip(zip(*self.argv), kwargList)]

        # Set up progress bar: the while loop ensures all futures are executed
        totalTasks = len(futures)
        pbar = tqdm.tqdm(total=totalTasks, bar_format=self.tqdmFormat, position=0, leave=True)
        cnt = 0
        while any(f.status == "pending" for f in futures):
            time.sleep(self.sleepTime)
            new = max(0, sum([f.status == "finished" for f in futures]) - cnt)
            cnt += new
            pbar.update(new)
        pbar.close()

        # Avoid race condition: give futures time to perform switch from 'pending'
        # to 'finished' so that `finishedTasks` is computed correctly
        log.debug("Waiting %f seconds for futures", self.sleepTime)
        time.sleep(self.sleepTime)

        # If number of 'finished' tasks is less than expected, go into
        # problem analysis mode: all futures that erred hav an `.exception`
        # method which can be used to track down the worker it was executed by
        # Once we know the worker, we can point to the right log file. If
        # futures were cancelled (by the user or the SLURM controller),
        # `.exception` is `None` and we can't reliably track down the
        # respective executing worker
        finishedTasks = sum([f.status == "finished" for f in futures])
        if finishedTasks < totalTasks:
            schedulerLog = list(self.client.cluster.get_logs(cluster=False, scheduler=True, workers=False).values())[0]
            erredFutures = [f for f in futures if f.status == "error"]
            msg = "%s Parallel computation failed: %d/%d tasks failed or stalled. "
            msg = msg%(self.objName, totalTasks - finishedTasks, totalTasks)
            msg += "Concurrent computing scheduler log info: "
            msg += schedulerLog + "\n"

            # If we're working w/`SLURMCluster`, perform the Herculean task of
            # tracking down which dask worker was executed by which SLURM job...
            if self.client.cluster.__class__.__name__ == "SLURMCluster":
                try:
                    erredJobs = [f.exception().last_worker.identity()["id"] for f in erredFutures]
                except AttributeError:
                    erredJobs = []
                erredJobs = list(set(erredJobs))
                validIDs = [job for job in erredJobs if job in self.client.cluster.workers.keys()]
                erredJobIDs = [self.client.cluster.workers[job].job_id for job in validIDs]
                errFiles = glob.glob(logDir + os.sep + "*.err")
                if len(erredFutures) > 0 or len(errFiles) > 0:
                    msg += "Please consult the following SLURM log files for details:\n"
                    if len(erredJobIDs) > 0:
                        msg += "".join(logFiles.format(id) + "\n" for id in erredJobIDs)
                    else:
                        msg += "".join(logDir)
                    msg += "".join(errfile + "\n" for errfile in errFiles)
                else:
                    msg += "Please check SLURM logs in %s"%(logDir)

            # In case of a `LocalCluster`, syphon worker logs
            else:                                                       # pragma: no cover
                msg += "Parallel worker log details: \n"
                workerLogs = self.client.get_worker_logs().values()
                for wLog in workerLogs:
                    if "Failed" in wLog:
                        msg += wLog

            # Finally, raise an error and get outta here
            raise RuntimeError(msg)

        # Postprocessing of results
        values = self.post_process(futures)

        # Either return collected by-worker results or the filepaths of results
        return values

    def post_process(
            self,
            futures: dd.Future) -> Union[List, None]:
        """
        Local helper to post-process results on disk/in-memory

        The return `values` is either
        `None` : if neither in-memory results collection or auto-writing was requested
        list of file-names: if `write_worker_results` is `True`
        list of objects: if in-memory results collection was requested
        """

        # Deduce result output information
        write_worker_results = self.acme_func == self.func_wrapper
        single_file = False
        if write_worker_results:
            write_pickle = self.results_container is None
            if not write_pickle and self.kwargv.get("singleFile") is not None:
                single_file = True
        else:
            write_pickle = False
        msg = "Inferred that `write_worker_results = %s`, `single_file = %s`, `write_pickle = %s`"
        log.debug(msg, str(write_worker_results), str(single_file), str(write_pickle))

        # If wanted (not recommended) collect computed results in local memory
        if self.collect_results:
            if not isSpyModule:
                log.info("Gathering results in local memory")
            collected = self.client.gather(futures)
            log.debug("Gathered results from client in a %d-element list", len(collected))
            if self.result_shape is not None:
                log.debug("Returning single NumPy array of shape %s and type %s",
                          str(self.result_shape), str(self.result_dtype))
                values = []
                arrVal = np.empty(shape=self.result_shape, dtype=self.result_dtype)
                idx = [slice(None)] * len(self.result_shape)
                for i, res in enumerate(collected):
                    if not isinstance(res, (list, tuple)):
                        res = [res]
                    idx[self.stacking_dim] = i
                    arrVal[tuple(idx)] = res[0]
                    for r in res[1:]:
                        values.append(r)
                values.insert(0, arrVal)
                # If `values` is a single array, don't wrap it inside a list
                if len(values) == 1:
                    values = values[0]
            else:
                log.debug("Returning a list of values")
                values = collected
        else:
            values = None

        # Prepare final output message
        successMsg = "SUCCESS!"

        # If automatic results writing was requested, perform some housekeeping
        if write_worker_results:
            finalMsg = "Results have been saved to %s"
            if write_pickle:
                log.debug("Saved results as pickle files")
                values = list(self.kwargv["outFile"])
                finalMsg = finalMsg%(self.out_dir)
                log.debug("Returning a list of file-names")
            else:
                if single_file:
                    log.debug("Saved results to single shared container")
                    finalMsg = finalMsg%(self.results_container)
                    if values is None:
                        values = [self.results_container]
                        log.debug("Returning container name as single-element list")
                else:
                    log.debug("Scanning payload directory for emergency pickles")
                    picklesFound = False
                    values = []
                    for fname in self.kwargv["outFile"]:
                        pklName = fname.rstrip(".h5") + ".pickle"
                        if os.path.isfile(fname):
                            values.append(fname)
                        elif os.path.isfile(pklName):
                            values.append(pklName)
                            picklesFound = True
                            log.debug("Found emergency pickle %s", pklName)
                        else:
                            missing = fname.rstrip(".h5")
                            values.append("Missing %s"%(missing))
                            log.debug("Missing file %s", missing)
                    payloadDir = os.path.dirname(values[0])

                    # If pickles are found, remove global `results_container` as it
                    # would contain invalid file-links and move compute results out
                    # of payload dir
                    if picklesFound:
                        os.unlink(self.results_container)               # type: ignore
                        wrng = "Some compute runs could not be saved as HDF5, " +\
                            "collection container %s has been removed as it would " +\
                                "comprise invalid file-links"
                        log.warning(wrng, self.results_container)
                        self.results_container = None

                        # Move files out of payload dir and update return `values`
                        target = os.path.abspath(os.path.join(payloadDir, os.pardir))
                        for i, fname in enumerate(values):
                            shutil.move(fname, target)
                            self.kwargv["outFile"][i] = os.path.join(target, os.path.basename(fname))
                            log.debug("Moved %s to %s", fname, target)
                        values = list(self.kwargv["outFile"])
                        log.debug("Returning a list of file-names")
                        shutil.rmtree(payloadDir)
                        log.debug("Deleted payload directory %s", payloadDir)
                        successMsg = ""
                        finalMsg = finalMsg%(target)

                    # All good, no pickle gymnastics was needed
                    else:

                        # In case of multiple return values present in by-worker
                        # containers but missing in collection container (happens
                        # if `result_shape` is not `None` and data-sets have to
                        # be pre-allocated), create "symlinks" to corresponding
                        # missing returns
                        log.debug("No emergency pickles found")
                        if self.stacking_dim is not None:
                            msg = "Check if additional return values " +\
                                "need to be added to container with pre-allocated dataset"
                            log.debug(msg)
                            with h5py.File(self.results_container, "r") as h5r:
                                with h5py.File(values[0], "r") as h5Tmp:
                                    missingReturns = set(h5Tmp.keys()).difference(h5r.keys())
                            if len(missingReturns) > 0:
                                log.debug("Found return values to be added")
                                with h5py.File(self.results_container, "a") as h5r:
                                    for retVal in missingReturns:
                                        for i, fname in enumerate(values):
                                            relPath = os.path.join(os.path.basename(payloadDir), os.path.basename(fname))
                                            h5r["comp_{}/{}".format(i, retVal)] = h5py.ExternalLink(relPath, retVal)
                                            log.debug("Added return value via external link comp_%d/%s", i, retVal)

                        finalMsg = finalMsg%(self.results_container)
                        msg = "Container ready, links to data payload located in %s"
                        log.debug(msg, payloadDir)
                        log.debug("Returning a list of file-names")
        else:
            finalMsg = "Finished parallel computation"

        # Print final triumphant output message and force-flush all logging handlers
        if len(successMsg) > 0:
            log.announce(successMsg)                                    # type: ignore
        log.info(finalMsg)
        for h in log.handlers:
            if hasattr(h, "flush"):
                h.flush()
        return values

    def cleanup(self) -> None:
        """
        Shut down any ad-hoc distributed computing clients created by `prepare_client`
        """

        # If `prepare_client` has not been launched yet, just get outta here
        if not hasattr(self, "client"):                                 # pragma: no cover
            log.debug("Helper `prepare_client` not yet launched, exiting")
            return
        if self.stop_client and self.client is not None:
            log.debug("Found client %s, calling `cluster_cleanup`", str(self.client))
            cluster_cleanup(self.client)
            self.client = None
            return
        log.debug("Either `stop_client = False` or no client found, returning")

        return

    @staticmethod
    def func_wrapper(*args: Any, **kwargs: Optional[Any]) -> None:      # pragma: no cover
        """
        If the output of `func` is saved to disk, wrap `func` with this static
        method to take care of filling up HDF5/pickle files

        If writing to HDF5 fails, use an "emergency-pickling" mechanism to try
        to save the output of `func` using pickle instead
        """

        # Extract everything from `kwargs` appended by `ACMEdaemon`
        func = kwargs.pop("userFunc")
        taskID = kwargs.pop("taskID")
        fname = kwargs.pop("outFile")
        logName = kwargs.pop("logName")
        logLevel = kwargs.pop("logLevel")
        singleFile = kwargs.pop("singleFile", False)
        stackingDim = kwargs.pop("stackingDim", None)
        memEstRun = kwargs.pop("memEstRun", False)

        # Set up logger
        log = logging.getLogger(logName)
        log.setLevel(logLevel)                                          # type: ignore
        for h in log.handlers:
            h.setLevel(logLevel)                                        # type: ignore

        # Call user-provided function
        result = func(*args, **kwargs)                                  # type: ignore

        # For memory estimation runs, don't start saving stuff
        if memEstRun:
            return

        # Save results: either (try to) use HDF5 or pickle stuff
        if fname.endswith(".h5"):                                       # type: ignore

            grpName = ""
            if singleFile:
                lock = dd.lock.Lock(name=os.path.basename(fname))       # type: ignore
                lock.acquire()
                grpName = "comp_{}/".format(taskID)

            if not isinstance(result, (list, tuple)):
                result = [result]

            try:

                with h5py.File(fname, "a") as h5f:
                    if stackingDim is None:
                        if not all(isinstance(value, (numbers.Number, str)) for value in result):
                            for rk, res in enumerate(result):
                                h5f.create_dataset(grpName + "result_{}".format(rk), data=res)
                                log.debug("Created new dataset `result_%d` in %s", rk, fname)
                        else:
                            h5f.create_dataset(grpName + "result_0", data=result)
                            log.debug("Created new dataset `result_0` in %s", fname)
                    else:
                        if singleFile:
                            idx = [slice(None)] * len(h5f["result_0"].shape)    # type: ignore
                            idx[stackingDim] = taskID                           # type: ignore
                            h5f["result_0"][tuple(idx)] = result[0]
                            log.debug("Wrote to pre-allocated dataset `result_0` in %s", fname)
                            for rk, res in enumerate(result[1:]):
                                h5f.create_dataset(grpName + "result_{}".format(rk + 1), data=res)
                                log.debug("Created new dataset `result_%d` in %s", rk+1, fname)
                        else:
                            for rk, res in enumerate(result):
                                h5f.create_dataset(grpName + "result_{}".format(rk), data=res)
                                log.debug("Created new dataset `result_%d` in %s", rk, fname)

                if singleFile:
                    lock.release()

            except TypeError as exc:

                if ("has no native HDF5 equivalent" in str(exc) \
                    or "One of data, shape or dtype must be specified" in str(exc)) \
                        and not singleFile:
                    try:
                        os.unlink(fname)                                # type: ignore
                        pname = fname.rstrip(".h5") + ".pickle"         # type: ignore
                        with open(os.path.join(pname), "wb") as pkf:
                            pickle.dump(result, pkf)
                        msg = "Could not write %s results have been pickled instead: %s. Return values are most likely " +\
                            "not suitable for storage in HDF5 containers. Original error message: %s"
                        log.warning(msg, fname, pname, str(exc))
                    except pickle.PicklingError as pexc:
                        err = "Unable to write %s, successive attempts to pickle results failed too: %s"
                        log.error(err, fname, str(pexc))
                else:
                    if singleFile:
                        err = "Could not write to %s. File potentially corrupted. Original error message: %s"
                        lock.release()
                    else:
                        err = "Could not access %s. Original error message: %s"
                    log.error(err, fname, str(exc))
                    raise exc

            except Exception as exc:

                if singleFile:
                    lock.release()
                raise exc

        else:

            try:
                with open(os.path.join(fname), "wb") as pkf:            # type: ignore
                    pickle.dump(result, pkf)
                    log.debug("Pickled to %s", fname)
            except pickle.PicklingError as pexc:
                err = "Could not pickle results to file %s. Original error message: %s"
                log.error(err, fname, str(pexc))
                raise pexc

        return
